"""Keywords in the Greenplum Dialect.

https://docs.vmware.com/en/VMware-Greenplum/7/greenplum-database/ref_guide-sql-keywords.html
"""

greenplum_keywords = [
    ("ACTIVE", "non-reserved"),
    ("CONCURRENCY", "non-reserved"),
    ("CONTAINS", "non-reserved"),
    ("COORDINATOR", "non-reserved"),
    ("CPU_RATE_LIMIT", "non-reserved"),
    ("CPUSET", "non-reserved"),
    ("CREATEEXTTABLE", "non-reserved"),
    ("CUBE", "non-reserved-(cannot-be-function-or-type)"),
    ("DECODE", "reserved"),
    ("DENY", "non-reserved"),
    ("DISTRIBUTED", "reserved"),
    ("DXL", "non-reserved"),
    ("ERRORS", "non-reserved"),
    ("EVERY", "non-reserved"),
    ("EXCHANGE", "non-reserved"),
    # ("EXCLUDE", "reserved"),
    ("EXPAND", "non-reserved"),
    ("FIELDS", "non-reserved"),
    ("FILL", "non-reserved"),
    # ("FOLLOWING", "reserved"),
    ("FORMAT", "non-reserved"),
    ("FULLSCAN", "non-reserved"),
    ("GROUP_ID", "non-reserved-(cannot-be-function-or-type)"),
    ("GROUPING", "non-reserved-(cannot-be-function-or-type)"),
    ("HASH", "non-reserved"),
    ("HOST", "non-reserved"),
    ("IGNORE", "non-reserved"),
    ("INCLUSIVE", "non-reserved"),
    ("LIST", "non-reserved"),
    ("LOG", "reserved-(can-be-function-or-type)"),
    ("MASTER", "non-reserved"),
    ("MEDIAN", "non-reserved-(cannot-be-function-or-type)"),
    ("MEMORY_LIMIT", "non-reserved"),
    ("MEMORY_SHARED_QUOTA", "non-reserved"),
    ("MEMORY_SPILL_RATIO", "non-reserved"),
    ("MISSING", "non-reserved"),
    ("MODIFIES", "non-reserved"),
    ("NEWLINE", "non-reserved"),
    ("NOCREATEEXTTABLE", "non-reserved"),
    ("NOOVERCOMMIT", "non-reserved"),
    ("ORDERED", "non-reserved"),
    ("OTHERS", "non-reserved"),
    ("OVERCOMMIT", "non-reserved"),
    # ("PARTITION", "reserved"),
    ("PARTITIONS", "non-reserved"),
    ("PERCENT", "non-reserved"),
    ("PERSISTENTLY", "non-reserved"),
    # ("PRECEDING", "reserved"),
    ("PROTOCOL", "non-reserved"),
    ("QUEUE", "non-reserved"),
    ("RANDOMLY", "non-reserved"),
    ("READABLE", "non-reserved"),
    ("READS", "non-reserved"),
    ("REJECT", "non-reserved"),
    ("REPLICATED", "non-reserved"),
    ("RESOURCE", "non-reserved"),
    ("ROLLUP", "non-reserved-(cannot-be-function-or-type)"),
    ("ROOTPARTITION", "non-reserved"),
    ("SCATTER", "reserved"),
    ("SEGMENT", "non-reserved"),
    ("SEGMENTS", "non-reserved"),
    ("SETS", "non-reserved-(cannot-be-function-or-type)"),
    ("SPLIT", "non-reserved"),
    ("SQL", "non-reserved"),
    ("SUBPARTITION", "non-reserved"),
    ("THRESHOLD", "non-reserved"),
    ("TIES", "non-reserved"),
    # ("UNBOUNDED", "reserved"),
    ("VALIDATION", "non-reserved"),
    ("WEB", "non-reserved"),
    ("WRITABLE", "non-reserved"),
]



"""Defines the jinja builtins for dbt."""

from typing import Any, Callable, Union

from sqlfluff.core.templaters.builtins.common import FunctionWrapper


class RelationEmulator:
    """A class which emulates the `this` class from dbt."""

    # Tell Jinja this object is safe to call and does not alter data.
    # https://jinja.palletsprojects.com/en/3.0.x/sandbox/#jinja2.sandbox.SandboxedEnvironment.is_safe_callable
    unsafe_callable = False
    alters_data = False

    identifier = "this_model"
    schema = "this_schema"
    database = "this_database"

    def __init__(self, identifier: str = "this_model") -> None:
        self.identifier = identifier

    def __call__(self, *args: Any, **kwargs: Any) -> "RelationEmulator":
        """When relation(*) is called return self as another relation."""
        return self

    def __getattr__(self, name: str) -> Union["RelationEmulator", bool]:
        """When relation.attribute is called return self as another relation.

        NOTE: If the attribute begins with `is_`, then return a boolean True.
        """
        if name[0:3] == "is_":
            return True
        return self

    def __str__(self) -> str:
        return self.identifier


class VarEmulator(str):
    """A string-like class which emulates dbt's `var()` return value."""

    def __new__(cls, value: str = "item") -> "VarEmulator":
        """Construct a string-like placeholder for mocked dbt `var()` output."""
        return super().__new__(cls, value)

    def __getattr__(self, name: str) -> "VarEmulator":
        """When var.attribute is called return self as another placeholder."""
        return self

    def __getitem__(self, name: Any) -> str:
        """When var['key'] is called return a placeholder instead of failing."""
        if isinstance(name, (int, slice)):
            return super().__getitem__(name)
        return self


class MacroReturn(Exception):
    """Exception used to allow macros to return non-string values."""

    def __init__(self, value: Any) -> None:
        self.value = value


def _builtin_return(value: Any) -> None:
    raise MacroReturn(value)


class DbtMacroWrapper:
    """Class that wraps macros to catch MacroReturn exceptions during execution."""

    # Tell Jinja this object is safe to call and does not alter data.
    # https://jinja.palletsprojects.com/en/3.0.x/sandbox/#jinja2.sandbox.SandboxedEnvironment.is_safe_callable
    unsafe_callable = False
    alters_data = False

    def __init__(self, macro: Callable[..., Any]):
        self.macro = macro

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Wraps jinja macro calls to handle dbt's `return` jinja function."""
        try:
            return self.macro(*args, **kwargs)
        except MacroReturn as e:
            return e.value


# NOTE: we use `FunctionWrapper` on all of the callable builtins here
# so that there's a sensible error message if someone tries to render
# them directly.
DBT_BUILTINS = {
    "ref": FunctionWrapper("ref", lambda *args, **kwargs: RelationEmulator(args[-1])),
    # In case of a cross project ref in dbt, model_ref is the second
    # argument. Otherwise it is the only argument.
    "source": FunctionWrapper(
        "source",
        lambda source_name, table: RelationEmulator(f"{source_name}_{table}"),
    ),
    "function": FunctionWrapper(
        "function",
        # Mirror `ref` handling: tolerate optional qualifier args and use the
        # last positional argument as the function identifier.
        lambda *args, **kwargs: RelationEmulator(args[-1]),
    ),
    "config": FunctionWrapper("config", lambda **kwargs: ""),
    "var": FunctionWrapper("var", lambda variable, default="": VarEmulator()),
    # `is_incremental()` renders as True, always in this case.
    # TODO: This means we'll never parse other parts of the query,
    # that are only reachable when `is_incremental()` returns False.
    # We should try to find a solution to that. Perhaps forcing the file
    # to be parsed TWICE if it uses this variable.
    "is_incremental": FunctionWrapper("is_incremental", lambda: True),
    "this": RelationEmulator(),
    "zip_strict": FunctionWrapper("zip_strict", zip),
    "zip": FunctionWrapper(
        "zip",
        lambda *args, default=None: (
            zip(*args) if all(hasattr(arg, "__iter__") for arg in args) else default
        ),
    ),
    "return": FunctionWrapper("return", _builtin_return),
}



"""A collection of decorators to modify rule docstrings for Sphinx.

NOTE: All of these decorators are deprecated from SQLFluff 2.0.0 onwards.

They are still included to allow a transition period, but the functionality
is now packaged in the BaseRule class via the RuleMetaclass.
"""

from typing import TYPE_CHECKING, Any

from sqlfluff.core.rules.base import rules_logger

if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.rules.base import BaseRule


def document_fix_compatible(cls: type["BaseRule"]) -> type["BaseRule"]:
    """Mark the rule as fixable in the documentation."""
    rules_logger.warning(
        f"{cls.__name__} uses the @document_fix_compatible decorator "
        "which is deprecated in SQLFluff 2.0.0. Remove the decorator "
        "to resolve this warning."
    )
    return cls


def document_groups(cls: type["BaseRule"]) -> type["BaseRule"]:
    """Mark the rule as fixable in the documentation."""
    rules_logger.warning(
        f"{cls.__name__} uses the @document_groups decorator "
        "which is deprecated in SQLFluff 2.0.0. Remove the decorator "
        "to resolve this warning."
    )
    return cls


def document_configuration(cls: type["BaseRule"], **kwargs: Any) -> type["BaseRule"]:
    """Add a 'Configuration' section to a Rule docstring."""
    rules_logger.warning(
        f"{cls.__name__} uses the @document_configuration decorator "
        "which is deprecated in SQLFluff 2.0.0. Remove the decorator "
        "to resolve this warning."
    )
    return cls



"""Implementation of Rule CV11."""

from collections.abc import Iterable
from typing import Optional

from sqlfluff.core.parser import (
    BaseSegment,
    KeywordSegment,
    SymbolSegment,
    WhitespaceSegment,
    WordSegment,
)
from sqlfluff.core.parser.segments.bracketed import BracketedSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.dialects.dialect_ansi import (
    FunctionContentsSegment,
    FunctionNameSegment,
    FunctionSegment,
)
from sqlfluff.utils.functional import FunctionalContext, Segments, sp


class Rule_CV11(BaseRule):
    """Enforce consistent type casting style.

    .. note::
        This is only compatible with 2-arguments CONVERT as
        some dialects allow an optional 3rd argument e.g TSQL,
        which cannot be rewritten into CAST.
        This rule is disabled by default for Teradata because it supports different
        type casting apart from CONVERT and ::
        e.g DATE '2007-01-01', '9999-12-31' (DATE).

    **Anti-pattern**

    Using mixture of CONVERT, :: and CAST when ``preferred_type_casting_style``
    config is set to ``consistent`` (default).

    .. code-block:: sql

        SELECT
            CONVERT(int, 1) AS bar,
            100::int::text,
            CAST(10 AS text) AS coo
        FROM foo;

    **Best practice**

    Use consistent type casting style.

    .. code-block:: sql

        SELECT
            CAST(1 AS int) AS bar,
            CAST(CAST(100 AS int) AS text),
            CAST(10 AS text) AS coo
        FROM foo;

    """

    name = "convention.casting_style"
    aliases = ("L067",)
    groups = ("all", "convention")
    config_keywords = ["preferred_type_casting_style"]
    crawl_behaviour = SegmentSeekerCrawler({"function", "cast_expression"})
    is_fix_compatible = True

    @staticmethod
    def _get_children(segments: Segments) -> Segments:
        return segments.children(
            sp.and_(
                sp.not_(sp.is_meta()),
                sp.not_(
                    sp.is_type(
                        "start_bracket",
                        "end_bracket",
                        "whitespace",
                        "newline",
                        "casting_operator",
                        "comma",
                    )
                ),
                # Exclude keywords, except when the keyword is a literal
                # (e.g. null_literal, boolean_literal).
                sp.or_(
                    sp.not_(sp.is_type("keyword")),
                    sp.is_type("literal"),
                ),
            )
        )

    @staticmethod
    def _build_function(name: str, contents: list[BaseSegment]) -> FunctionSegment:
        """Construct a parse-shaped ``function`` segment.

        The nesting (``function > function_name`` and ``function >
        function_contents > bracketed``) must mirror how the same SQL would
        parse. The fix loop doesn't reparse between rule passes, so a flat
        sequence of raw tokens here would present a different shape to
        later rules in the same pass — reflow spacing (LT01) keys "touch"
        off the ``function_name``/``function_contents`` containers and would
        wrongly insert a space after the function name (which the next,
        freshly-parsed run would then remove again: a non-convergent fix).
        """
        start = SymbolSegment("(", type="start_bracket")
        end = SymbolSegment(")", type="end_bracket")
        return FunctionSegment(
            segments=(
                FunctionNameSegment(
                    segments=(WordSegment(name, type="function_name_identifier"),)
                ),
                FunctionContentsSegment(
                    segments=(
                        BracketedSegment(
                            segments=(start, *contents, end),
                            start_bracket=(start,),
                            end_bracket=(end,),
                        ),
                    )
                ),
            )
        )

    @classmethod
    def _cast_fix_list(
        cls,
        context: RuleContext,
        cast_arg_1: Iterable[BaseSegment],
        cast_arg_2: BaseSegment,
        later_types: Optional[Segments] = None,
    ) -> list[LintFix]:
        """Generate list of fixes to convert CONVERT and ShorthandCast to CAST."""
        cast_function = cls._build_function(
            "cast",
            list(cast_arg_1)
            + [
                WhitespaceSegment(),
                KeywordSegment("as"),
                WhitespaceSegment(),
                cast_arg_2,
            ],
        )

        if later_types:
            for _type in later_types:
                cast_function = cls._build_function(
                    "cast",
                    [
                        cast_function,
                        WhitespaceSegment(),
                        KeywordSegment("as"),
                        WhitespaceSegment(),
                        _type,
                    ],
                )

        fixes = [
            LintFix.replace(
                context.segment,
                [cast_function],
            )
        ]
        return fixes

    @classmethod
    def _convert_fix_list(
        cls,
        context: RuleContext,
        convert_arg_1: BaseSegment,
        convert_arg_2: BaseSegment,
        later_types=None,
    ) -> list[LintFix]:
        """Generate list of fixes to convert CAST and ShorthandCast to CONVERT."""
        convert_function = cls._build_function(
            "convert",
            [
                convert_arg_1,
                SymbolSegment(",", type="comma"),
                WhitespaceSegment(),
                convert_arg_2,
            ],
        )

        if later_types:
            for _type in later_types:
                convert_function = cls._build_function(
                    "convert",
                    [
                        _type,
                        SymbolSegment(",", type="comma"),
                        WhitespaceSegment(),
                        convert_function,
                    ],
                )

        fixes = [
            LintFix.replace(
                context.segment,
                [convert_function],
            )
        ]
        return fixes

    @staticmethod
    def _shorthand_fix_list(
        context: RuleContext, shorthand_arg_1: BaseSegment, shorthand_arg_2: BaseSegment
    ) -> list[LintFix]:
        """Generate list of fixes to convert CAST and CONVERT to ShorthandCast."""
        if len(shorthand_arg_1.raw_segments) > 1:
            edits = [
                SymbolSegment("(", type="start_bracket"),
                shorthand_arg_1,
                SymbolSegment(")", type="end_bracket"),
            ]
        else:
            edits = [shorthand_arg_1]
        edits.extend(
            [
                SymbolSegment("::", type="casting_operator"),
                shorthand_arg_2,
            ]
        )

        fixes = [
            LintFix.replace(
                context.segment,
                edits,
            )
        ]
        return fixes

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Enforce consistent type casting style."""
        # Config type hints
        self.preferred_type_casting_style: str

        # Rule disabled for teradata, athena, or trino.
        # They only support one of the options
        # TODO: add additional dialects that only support a single option here.
        # TODO: add a tier list for dialects support multiples, but not all.
        if context.dialect.name in ("teradata", "athena", "trino"):
            return None

        # If we're in a templated section, don't consider the current location.
        # (i.e. if a cast happens in a macro, the end user writing the current
        # query may not know that or have control over it, so we should just
        # skip it).
        if context.segment.pos_marker:
            if not context.segment.pos_marker.is_literal():
                return None

        # Construct segment type casting
        if context.segment.is_type("function"):
            function_name = context.segment.get_child("function_name")
            # Functions should always have a name, that means this clause should
            # be unnecessary.
            if not function_name:  # pragma: no cover
                return None
            elif function_name.raw_upper == "CAST":
                current_type_casting_style = "cast"
            elif function_name.raw_upper == "CONVERT":
                current_type_casting_style = "convert"
            else:
                current_type_casting_style = None
        elif context.segment.is_type("cast_expression"):
            # CV11 can only rewrite the standard `::` shorthand. Dialect-specific
            # variants such as Databricks `?::` (try_cast) and Vertica `::!`
            # (null cast) are not interchangeable with CAST/CONVERT, so rewriting
            # them would change semantics (and may even be unparsable). Only treat
            # the expression as a shorthand cast when it uses the `::` operator.
            if not context.segment.get_child("casting_operator"):
                return None
            current_type_casting_style = "shorthand"
        else:  # pragma: no cover
            current_type_casting_style = None

        functional_context = FunctionalContext(context)

        # If casting style is set to consistent,
        # we use the casting style of the first segment we encounter.
        # convert_content = None
        if self.preferred_type_casting_style == "consistent":
            memory = context.memory
            prior_type_casting_style = context.memory.get("prior_type_casting_style")
            previous_skipped = context.memory.get("previous_skipped")

            # if previous_skipped then we can skip the whole fix
            # Construct fixes
            if prior_type_casting_style == "cast":
                if current_type_casting_style == "convert":
                    # Get the content of CONVERT
                    bracketed = functional_context.segment.children(
                        sp.is_type("function_contents")
                    ).children(sp.is_type("bracketed"))
                    convert_content = self._get_children(bracketed)
                    # We only care about 2-arguments convert
                    # some dialects allow an optional 3rd argument e.g TSQL
                    # which cannot be rewritten into CAST
                    if len(convert_content) > 2:
                        # set previous_skipped
                        if previous_skipped is None:
                            # Only update prior_type_casting_style
                            # if it is none, this ultimately
                            # makes sure we maintain the first
                            # casting style we encounter
                            memory["previous_skipped"] = True
                        return None

                    fixes = self._cast_fix_list(
                        context,
                        [convert_content[1]],
                        convert_content[0],
                    )
                elif current_type_casting_style == "shorthand":
                    # Get the expression and the datatype segment
                    expression_datatype_segment = self._get_children(
                        functional_context.segment
                    )

                    fixes = self._cast_fix_list(
                        context,
                        [expression_datatype_segment[0]],
                        expression_datatype_segment[1],
                        # We can have multiple shorthandcast e.g 1::int::text
                        # in that case, we need to introduce nested CAST()
                        expression_datatype_segment[2:],
                    )

            elif prior_type_casting_style == "convert":
                bracketed = functional_context.segment.children(
                    sp.is_type("function_contents")
                ).children(sp.is_type("bracketed"))
                if current_type_casting_style == "cast":
                    cast_content = self._get_children(bracketed)
                    if len(cast_content) > 2:
                        return None

                    fixes = self._convert_fix_list(
                        context,
                        cast_content[1],
                        cast_content[0],
                    )
                elif current_type_casting_style == "shorthand":
                    expression_datatype_segment = self._get_children(
                        functional_context.segment
                    )
                    fixes = self._convert_fix_list(
                        context,
                        expression_datatype_segment[1],
                        expression_datatype_segment[0],
                        expression_datatype_segment[2:],
                    )
            elif prior_type_casting_style == "shorthand":
                bracketed = functional_context.segment.children(
                    sp.is_type("function_contents")
                ).children(sp.is_type("bracketed"))
                if current_type_casting_style == "cast":
                    # Get the content of CAST
                    cast_content = self._get_children(bracketed)
                    if len(cast_content) > 2:
                        return None

                    fixes = self._shorthand_fix_list(
                        context,
                        cast_content[0],
                        cast_content[1],
                    )
                elif current_type_casting_style == "convert":
                    convert_content = self._get_children(bracketed)
                    if len(convert_content) > 2:
                        return None

                    fixes = self._shorthand_fix_list(
                        context,
                        convert_content[1],
                        convert_content[0],
                    )

            if (
                prior_type_casting_style
                and current_type_casting_style
                and (prior_type_casting_style != current_type_casting_style)
            ):
                return LintResult(
                    anchor=context.segment,
                    memory=context.memory,
                    fixes=fixes,
                    description=("Inconsistent type casting styles found."),
                )

            if prior_type_casting_style is None:
                # Only update prior_type_casting_style if it is none, this ultimately
                # makes sure we maintain the first casting style we encounter
                memory["prior_type_casting_style"] = current_type_casting_style
        elif (
            current_type_casting_style
            and current_type_casting_style != self.preferred_type_casting_style
        ):
            convert_content = None
            cast_content = None
            if self.preferred_type_casting_style == "cast":
                if current_type_casting_style == "convert":
                    bracketed = functional_context.segment.children(
                        sp.is_type("function_contents")
                    ).children(sp.is_type("bracketed"))
                    convert_content = self._get_children(bracketed)

                    fixes = self._cast_fix_list(
                        context,
                        [convert_content[1]],
                        convert_content[0],
                    )
                elif current_type_casting_style == "shorthand":
                    expression_datatype_segment = self._get_children(
                        functional_context.segment
                    )

                    for data_type_idx, seg in enumerate(expression_datatype_segment):
                        if seg.is_type("data_type"):
                            break

                    fixes = self._cast_fix_list(
                        context,
                        expression_datatype_segment[:data_type_idx],
                        expression_datatype_segment[data_type_idx],
                        expression_datatype_segment[data_type_idx + 1 :],
                    )

            elif self.preferred_type_casting_style == "convert":
                if current_type_casting_style == "cast":
                    bracketed = functional_context.segment.children(
                        sp.is_type("function_contents")
                    ).children(sp.is_type("bracketed"))
                    cast_content = self._get_children(bracketed)
                    fixes = self._convert_fix_list(
                        context,
                        cast_content[1],
                        cast_content[0],
                    )
                elif current_type_casting_style == "shorthand":
                    expression_datatype_segment = self._get_children(
                        functional_context.segment
                    )
                    fixes = self._convert_fix_list(
                        context,
                        expression_datatype_segment[1],
                        expression_datatype_segment[0],
                        expression_datatype_segment[2:],
                    )
            elif self.preferred_type_casting_style == "shorthand":
                bracketed = functional_context.segment.children(
                    sp.is_type("function_contents")
                ).children(sp.is_type("bracketed"))
                if current_type_casting_style == "cast":
                    cast_content = self._get_children(bracketed)
                    fixes = self._shorthand_fix_list(
                        context,
                        cast_content[0],
                        cast_content[1],
                    )
                elif current_type_casting_style == "convert":
                    convert_content = self._get_children(bracketed)
                    fixes = self._shorthand_fix_list(
                        context,
                        convert_content[1],
                        convert_content[0],
                    )

            # Don't fix if there's too much content.
            if (convert_content and len(convert_content) > 2) or (
                cast_content and len(cast_content) > 2
            ):
                fixes = []

            return LintResult(
                anchor=context.segment,
                memory=context.memory,
                fixes=fixes,
                description=(
                    "Used type casting style is different from"
                    " the preferred type casting style."
                ),
            )
        return None



"""Defines the formatters for the CLI."""

import os
import sys
from io import StringIO
from typing import Callable, Optional, Union

import click
from colorama import Style

from sqlfluff.cli import EXIT_FAIL, EXIT_SUCCESS
from sqlfluff.cli.helpers import (
    get_package_version,
    get_python_implementation,
    get_python_version,
    pad_line,
    wrap_field,
)
from sqlfluff.cli.outputstream import OutputKind, OutputPolicy, OutputStream
from sqlfluff.core import FluffConfig, Linter, SQLBaseError, TimingSummary
from sqlfluff.core.linter import FormatterInterface, LintedFile, ParsedString
from sqlfluff.core.types import Color


def split_string_on_spaces(s: str, line_length: int = 100) -> list[str]:
    """Split a string into lines based on whitespace.

    For short strings the functionality is trivial.
    >>> split_string_on_spaces("abc")
    ['abc']

    For longer sections it will split at an appropriate point.
    >>> split_string_on_spaces("abc def ghi", line_length=7)
    ['abc def', 'ghi']

    After splitting, multi-space sections should be intact.
    >>> split_string_on_spaces("a '   ' b c d e f", line_length=11)
    ["a '   ' b c", 'd e f']
    """
    line_buff = []
    str_buff = ""
    # NOTE: We *specify* the single space split, so that on reconstruction
    # we can accurately represent multi space strings.
    for token in s.split(" "):
        # Can we put this token on this line without going over?
        if str_buff:
            if len(str_buff) + len(token) > line_length:
                line_buff.append(str_buff)
                str_buff = token
            else:
                str_buff += " " + token
        else:
            # In the case that the buffer is already empty, add it without checking,
            # otherwise there might be things that we might never.
            str_buff = token
    # If we have left over buff, add it in
    if str_buff:
        line_buff.append(str_buff)
    return line_buff


def format_linting_result_header() -> str:
    """Format the header of a linting result output."""
    text_buffer = StringIO()
    text_buffer.write("==== readout ====\n")
    return text_buffer.getvalue()


class OutputStreamFormatter(FormatterInterface):
    """Formatter which writes to an OutputStream.

    On instantiation, this formatter accepts a function to
    dispatch messages. Each public method accepts an object
    or data in a common format, with this class handling the
    formatting and output.

    This class is designed to be subclassed if we eventually
    want to provide other methods of surfacing output.


    Args:
        output_stream: Output is sent here
        output_policy: Controls which semantic output categories are emitted
        filter_empty: If True, empty messages will not be dispatched
        output_line_length: Maximum line length
    """

    def __init__(
        self,
        output_stream: OutputStream,
        nocolor: Optional[bool],
        output_policy: OutputPolicy,
        filter_empty: bool = True,
        output_line_length: int = 80,
        show_lint_violations: bool = False,
    ):
        self._output_stream = output_stream
        self.plain_output = self.should_produce_plain_output(nocolor)
        self.output_policy = output_policy
        self._filter_empty = filter_empty
        self.output_line_length = output_line_length
        self.show_lint_violations = show_lint_violations

    @staticmethod
    def should_produce_plain_output(nocolor: Optional[bool]) -> bool:
        """Returns True if text output should be plain (not colored)."""
        if nocolor is False:
            return False
        if nocolor is True:
            return True
        return not sys.stdout.isatty() or bool(os.getenv("NO_COLOR"))

    def _dispatch(self, s: str) -> None:
        """Dispatch a string to the callback.

        This method is designed as a point for subclassing.
        """
        # The strip here is to filter out any empty messages
        if (not self._filter_empty) or s.strip(" \n\t"):
            self._output_stream.write(s)

    def dispatch_message(
        self,
        s: str,
        kind: OutputKind,
        *,
        minimum_verbosity: int = 0,
    ) -> None:
        """Dispatch a message when its semantic output category is enabled."""
        if self.output_policy.allows(kind, minimum_verbosity=minimum_verbosity):
            self._dispatch(s)

    def _format_config(self, linter: Linter) -> str:
        """Format the config of a `Linter`."""
        text_buffer = StringIO()
        if self.output_policy.allows(OutputKind.VERBOSE, minimum_verbosity=1):
            text_buffer.write("==== sqlfluff ====\n")
            config_content = [
                ("sqlfluff", get_package_version()),
                ("python", get_python_version()),
                ("implementation", get_python_implementation()),
                ("verbosity", self.output_policy.verbosity),
            ]
            if linter.dialect:
                config_content.append(("dialect", linter.dialect.name))
            config_content += linter.templater.config_pairs()
            text_buffer.write(
                self.cli_table(config_content, col_width=30, max_label_width=15)
            )
            text_buffer.write("\n")
            if linter.config.get("rule_allowlist"):
                text_buffer.write(
                    self.cli_table(
                        [("rules", ", ".join(linter.config.get("rule_allowlist")))],
                        col_width=41,
                    )
                )
            if self.output_policy.allows(OutputKind.VERBOSE, minimum_verbosity=2):
                text_buffer.write("\n== Raw Config:\n")
                text_buffer.write(self.format_config_vals(linter.config.iter_vals()))
        return text_buffer.getvalue()

    def dispatch_config(self, linter: Linter) -> None:
        """Dispatch configuration output appropriately."""
        self.dispatch_message(
            self._format_config(linter),
            OutputKind.VERBOSE,
            minimum_verbosity=1,
        )

    def dispatch_persist_filename(self, filename: str, result: str) -> None:
        """Dispatch filenames during a persist operation."""
        if result == "SKIP":
            self.dispatch_message(
                self.format_filename(filename=filename, success=result),
                OutputKind.VERBOSE,
                minimum_verbosity=2,
            )
        else:
            self.dispatch_message(
                self.format_filename(filename=filename, success=result),
                OutputKind.ACTION,
            )

    def _format_path(self, path: str) -> str:
        """Format paths."""
        return f"=== [ path: {self.colorize(path, Color.light)} ] ===\n"

    def dispatch_path(self, path: str) -> None:
        """Dispatch paths for display."""
        self.dispatch_message(
            self._format_path(path), OutputKind.VERBOSE, minimum_verbosity=1
        )

    def dispatch_template_header(
        self, fname: str, linter_config: FluffConfig, file_config: Optional[FluffConfig]
    ) -> None:
        """Dispatch the header displayed before templating."""
        if self.output_policy.allows(OutputKind.VERBOSE, minimum_verbosity=2):
            self.dispatch_message(
                self.format_filename(filename=fname, success="TEMPLATING"),
                OutputKind.VERBOSE,
                minimum_verbosity=2,
            )
            # This is where we output config diffs if they exist.
            if file_config:
                # Only output config diffs if there is a config to diff to.
                config_diff = file_config.diff_to(linter_config)
                if config_diff:  # pragma: no cover
                    self.dispatch_message(
                        "   Config Diff:",
                        OutputKind.VERBOSE,
                        minimum_verbosity=2,
                    )
                    self.dispatch_message(
                        self.format_config_vals(
                            linter_config.iter_vals(cfg=config_diff)
                        ),
                        OutputKind.VERBOSE,
                        minimum_verbosity=2,
                    )

    def dispatch_parse_header(self, fname: str) -> None:
        """Dispatch the header displayed before parsing."""
        self.dispatch_message(
            self.format_filename(filename=fname, success="PARSING"),
            OutputKind.VERBOSE,
            minimum_verbosity=2,
        )

    def dispatch_lint_header(self, fname: str, rules: list[str]) -> None:
        """Dispatch the header displayed before linting."""
        self.dispatch_message(
            self.format_filename(
                filename=fname, success=f"LINTING ({', '.join(rules)})"
            ),
            OutputKind.VERBOSE,
            minimum_verbosity=2,
        )

    def dispatch_compilation_header(self, templater: str, message: str) -> None:
        """Dispatch the header displayed before linting."""
        self.dispatch_message(
            f"=== [{self.colorize(templater, Color.light)}] {message}",
            OutputKind.STATUS,
        )  # pragma: no cover

    def dispatch_processing_header(self, processes: int) -> None:
        """Dispatch the header displayed before linting."""
        self.dispatch_message(  # pragma: no cover
            f"{self.colorize('effective configured processes: ', Color.light)} "
            f"{processes}",
            OutputKind.VERBOSE,
            minimum_verbosity=1,
        )

    def dispatch_dialect_warning(self, dialect: str) -> None:
        """Dispatch a warning for dialects."""
        self.dispatch_message(  # pragma: no cover
            self.format_dialect_warning(dialect), OutputKind.DIAGNOSTIC
        )

    def _format_file_violations(
        self, fname: str, violations: list[SQLBaseError]
    ) -> str:
        """Format a set of violations in a `LintingResult`."""
        text_buffer = StringIO()
        # Success is based on there being no fails, but we still
        # want to show the results if there are warnings (even
        # if no fails).
        fails = sum(
            int(not violation.ignore and not violation.warning)
            for violation in violations
        )
        warns = sum(int(violation.warning) for violation in violations)
        show = fails + warns > 0

        # Only print the filename if it has diagnostics or verbose output is enabled.
        if self.output_policy.allows(OutputKind.VERBOSE, minimum_verbosity=1) or show:
            text_buffer.write(self.format_filename(fname, success=fails == 0))
            text_buffer.write("\n")

        # If we have violations, print them
        if show:
            # sort by position in file (using line number and position)
            s = sorted(violations, key=lambda v: (v.line_no, v.line_pos))
            for violation in s:
                text_buffer.write(
                    self.format_violation(
                        violation, max_line_length=self.output_line_length
                    )
                )
                text_buffer.write("\n")
        str_buffer = text_buffer.getvalue()
        # Remove the trailing newline if there is one
        if len(str_buffer) > 0 and str_buffer[-1] == "\n":
            str_buffer = str_buffer[:-1]
        return str_buffer

    def dispatch_file_violations(
        self,
        fname: str,
        linted_file: LintedFile,
        only_fixable: bool,
        warn_unused_ignores: bool,
    ) -> None:
        """Dispatch any violations found in a file."""
        violations = linted_file.get_violations(
            fixable=(
                True if bool(only_fixable and not self.show_lint_violations) else None
            ),
            filter_warning=False,
            warn_unused_ignores=warn_unused_ignores,
        )
        s = self._format_file_violations(
            fname,
            violations,
        )
        kind = (
            OutputKind.STATUS
            if only_fixable and not self.show_lint_violations
            else OutputKind.DIAGNOSTIC
        )
        self.dispatch_message(s, kind)

    def colorize(self, s: str, color: Optional[Color] = None) -> str:
        """Optionally use ANSI colour codes to colour a string."""
        return self.colorize_helper(self.plain_output, s, color)

    @staticmethod
    def colorize_helper(
        plain_output: bool, s: str, color: Optional[Color] = None
    ) -> str:
        """Static version of colorize() method."""
        if not color or plain_output:
            return s
        else:
            return f"{color.value}{s}{Style.RESET_ALL}"

    def cli_table_row(
        self,
        fields: list[tuple[str, str]],
        col_width,
        max_label_width=10,
        sep_char=": ",
        divider_char=" ",
        label_color=Color.light,
        val_align="right",
    ) -> str:
        """Make a row of a CLI table, using wrapped values."""
        # Do some intel first
        cols = len(fields)
        last_col_idx = cols - 1
        wrapped_fields = [
            wrap_field(
                field[0],
                field[1],
                width=col_width,
                max_label_width=max_label_width,
                sep_char=sep_char,
            )
            for field in fields
        ]
        max_lines = max(fld["lines"] for fld in wrapped_fields)
        last_line_idx = max_lines - 1
        # Make some text
        buff = StringIO()
        for line_idx in range(max_lines):
            for col_idx in range(cols):
                # Assume we pad labels left and values right
                fld = wrapped_fields[col_idx]
                ll = fld["label_list"]
                vl = fld["val_list"]
                buff.write(
                    self.colorize(
                        pad_line(
                            ll[line_idx] if line_idx < len(ll) else "",
                            width=fld["label_width"],
                        ),
                        color=label_color,
                    )
                )
                if line_idx == 0:
                    buff.write(sep_char)
                else:
                    buff.write(" " * len(sep_char))
                buff.write(
                    pad_line(
                        vl[line_idx] if line_idx < len(vl) else "",
                        width=fld["val_width"],
                        align=val_align,
                    )
                )
                if col_idx != last_col_idx:
                    buff.write(divider_char)
                elif line_idx != last_line_idx:
                    buff.write("\n")
        return buff.getvalue()

    def cli_table(
        self,
        fields,
        col_width=20,
        cols=2,
        divider_char=" ",
        sep_char=": ",
        label_color=Color.light,
        float_format="{0:.2f}",
        max_label_width=10,
        val_align="right",
    ) -> str:
        """Make a crude ascii table.

        Assume that `fields` is an iterable of (label, value) pairs.
        """
        # First format all the values into strings
        formatted_fields = []
        for label, value in fields:
            label = str(label)
            if isinstance(value, float):
                value = float_format.format(value)
            else:
                value = str(value)
            formatted_fields.append((label, value))

        # Set up a buffer to hold the whole table
        buff = StringIO()
        while len(formatted_fields) > 0:
            row_buff: list[tuple[str, str]] = []
            while len(row_buff) < cols and len(formatted_fields) > 0:
                row_buff.append(formatted_fields.pop(0))
            buff.write(
                self.cli_table_row(
                    row_buff,
                    col_width=col_width,
                    max_label_width=max_label_width,
                    sep_char=sep_char,
                    divider_char=divider_char,
                    label_color=label_color,
                    val_align=val_align,
                )
            )
            if len(formatted_fields) > 0:
                buff.write("\n")
        return buff.getvalue()

    def format_filename(
        self,
        filename: str,
        success: Union[str, bool] = False,
        success_text: str = "PASS",
    ) -> str:
        """Format filenames."""
        if isinstance(success, str):
            status_string = success
        else:
            status_string = success_text if success else "FAIL"

        if status_string in ("PASS", "FIXED", success_text):
            status_string = self.colorize(status_string, Color.green)
        elif status_string in ("FAIL", "ERROR"):
            status_string = self.colorize(status_string, Color.red)

        return f"== [{self.colorize(filename, Color.light)}] {status_string}"

    def _format_line_and_pos(
        self, line_no: Optional[int], line_pos: Optional[int]
    ) -> tuple[str, str]:
        """Format line and position elements for consistent display.

        Args:
            line_no: Line number (None if not available)
            line_pos: Position within line (None if not available)

        Returns:
            tuple: (formatted_line_elem, formatted_pos_elem)
        """
        line_elem = "   -" if line_no is None else f"{line_no:4d}"
        pos_elem = "   -" if line_pos is None else f"{line_pos:4d}"
        return line_elem, pos_elem

    def format_violation(
        self,
        violation: Union[SQLBaseError, dict],
        max_line_length: int = 90,
    ) -> str:
        """Format a violation.

        NOTE: This method accepts both SQLBaseError objects and the serialised
        dict representation. If the former is passed, then the conversion is
        done within the method so we can work with a common representation.
        """
        if isinstance(violation, dict):
            v_dict: dict = violation
        elif isinstance(violation, SQLBaseError):
            v_dict = violation.to_dict()
        elif not isinstance(violation, dict):  # pragma: no cover
            raise ValueError(f"Unexpected violation format: {violation}")

        desc: str = v_dict["description"]
        code: str = v_dict["code"]
        name: str = v_dict["name"]
        line_no: int = v_dict["start_line_no"]
        line_pos: int = v_dict["start_line_pos"]
        warning: bool = v_dict["warning"]
        line_elem, pos_elem = self._format_line_and_pos(line_no, line_pos)

        if warning:
            desc = "WARNING: " + desc  # pragma: no cover

        # If the rule has a name, add that the description.
        if name:
            desc += f" [{self.colorize(name, Color.light)}]"

        split_desc = split_string_on_spaces(desc, line_length=max_line_length - 25)

        out_buff = ""
        # Grey out the violation if we're ignoring or warning it.
        section_color: Color
        if warning:
            section_color = Color.light
        else:
            section_color = Color.blue

        for idx, line in enumerate(split_desc):
            if idx == 0:
                rule_code = code.rjust(4)
                if "PRS" in rule_code:
                    section_color = Color.red
                out_buff += self.colorize(
                    f"L:{line_elem} | P:{pos_elem} | {rule_code} | ",
                    section_color,
                )
            else:
                out_buff += (
                    "\n"
                    + (" " * 23)
                    + self.colorize(
                        "| ",
                        section_color,
                    )
                )
            out_buff += line
        return out_buff

    def format_linting_stats(self, result) -> str:
        """Format a set of stats given a `LintingResult`."""
        text_buffer = StringIO()
        all_stats = result.stats(EXIT_FAIL, EXIT_SUCCESS)
        text_buffer.write("==== summary ====\n")
        if self.output_policy.allows(OutputKind.VERBOSE, minimum_verbosity=2):
            output_fields = [
                "files",
                "violations",
                "clean files",
                "unclean files",
                "avg per file",
                "unclean rate",
                "status",
            ]
            special_formats = {"unclean rate": "{0:.0%}"}
        else:
            output_fields = ["violations", "status"]
            special_formats = {}
        # Generate content tuples, applying special formats for some fields
        summary_content = [
            (
                key,
                (
                    special_formats[key].format(all_stats[key])
                    if key in special_formats
                    else all_stats[key]
                ),
            )
            for key in output_fields
        ]
        # Render it all as a table
        text_buffer.write(self.cli_table(summary_content, max_label_width=14))
        return text_buffer.getvalue()

    def format_config_vals(self, config_vals) -> str:
        """Format an iterable of config values from a config object."""
        text_buffer = StringIO()
        for i, k, v in config_vals:
            val = "" if v is None else str(v)
            text_buffer.write(
                ("    " * i)
                + self.colorize(pad_line(str(k) + ":", 20, "left"), color=Color.light)
                + pad_line(val, 20, "left")
                + "\n"
            )
        return text_buffer.getvalue()

    def _format_rule_description(self, rule) -> str:
        """Format individual rule.

        This is a helper function in .format_rules().
        """
        if rule.name:
            name = self.colorize(rule.name, Color.blue)
            description = f"[{name}] {rule.description}"
        else:
            description = rule.description

        if rule.groups:
            groups = self.colorize(", ".join(rule.groups), Color.light)
            description += f"\ngroups: {groups}"
        if rule.aliases:
            aliases = self.colorize(", ".join(rule.aliases), Color.light)
            description += f" aliases: {aliases}"
        return description

    def format_rules(self, linter: Linter, verbose: int = 0) -> str:
        """Format the a set of rules given a `Linter`."""
        text_buffer = StringIO()
        text_buffer.write("==== sqlfluff - rules ====\n")
        text_buffer.write(
            self.cli_table(
                [
                    (
                        t.code,
                        self._format_rule_description(t),
                    )
                    for t in linter.rule_tuples()
                ],
                col_width=80,
                cols=1,
                label_color=Color.blue,
                val_align="left",
            )
        )
        return text_buffer.getvalue()

    def format_dialects(self, dialect_readout, verbose=0) -> str:
        """Format the dialects yielded by `dialect_readout`."""
        text_buffer = StringIO()
        text_buffer.write("==== sqlfluff - dialects ====\n")
        readouts = [
            (
                dialect.label,
                f"{dialect.name} dialect [inherits from '{dialect.inherits_from}']",
            )
            for dialect in dialect_readout()
        ]
        text_buffer.write(
            self.cli_table(
                readouts,
                col_width=60,
                cols=1,
                label_color=Color.blue,
                val_align="right",
            )
        )
        return text_buffer.getvalue()

    def format_dialect_warning(self, dialect: str) -> str:
        """Output a warning for parsing errors."""
        return self.colorize(
            (
                "WARNING: Parsing errors found and dialect is set to "
                f"'{dialect}'. Have you configured your dialect correctly?"
            ),
            Color.light,
        )

    def print_out_residual_error_counts(
        self,
        total_errors: int,
        num_filtered_errors: int,
        tmp_prs_errors: Optional[dict[str, list[SQLBaseError]]] = None,
        force_stderr: bool = False,
    ) -> None:
        """Output the residual error totals for the file.

        Args:
            total_errors (int): The total number of templating & parsing errors.
            num_filtered_errors (int): The number of templating & parsing errors
                which remain after any noqa and filters applied.
            tmp_prs_errors (dict[str, list], optional): A dict mapping filenames to
                lists of templating/parsing errors for detailed reporting.
            force_stderr (bool): Whether to force the output onto stderr. By default
                the output is on stdout if there are no errors, otherwise stderr.
        """
        if total_errors and not self.show_lint_violations:
            click.echo(
                message=self.colorize(
                    f"  [{total_errors} templating/parsing errors found]", Color.red
                ),
                color=not self.plain_output,
                err=True,
            )

            # Show detailed error information for templating/parsing errors
            if tmp_prs_errors:
                # Errors are grouped by file
                for filepath, errors in tmp_prs_errors.items():
                    click.echo(
                        message=self.colorize(f"  == [{filepath}] ==", Color.light),
                        color=not self.plain_output,
                        err=True,
                    )
                    for error in errors:
                        formatted_error = self.format_violation(
                            violation=error, max_line_length=self.output_line_length
                        )
                        click.echo(
                            message=self.colorize(f"  {formatted_error}", Color.red),
                            color=not self.plain_output,
                            err=True,
                        )

            if num_filtered_errors < total_errors:
                color = Color.red if num_filtered_errors else Color.green
                click.echo(
                    message=self.colorize(
                        f"  [{num_filtered_errors} templating/parsing errors "
                        f'remaining after "ignore" & "warning"]',
                        color=color,
                    ),
                    color=not self.plain_output,
                    err=force_stderr or num_filtered_errors > 0,
                )

    def print_out_violations_and_timing(
        self,
        output_stream: OutputStream,
        bench: bool,
        code_only: bool,
        total_time: float,
        parsed_strings: list[ParsedString],
        violations_getter: Optional[
            Callable[[ParsedString], list[SQLBaseError]]
        ] = None,
    ) -> int:
        """Used by human formatting during the `sqlfluff parse` command."""
        violations_count = 0
        timing = TimingSummary()

        for parsed_string in parsed_strings:
            timing.add(parsed_string.time_dict)

            num_variants = len(parsed_string.parsed_variants)
            root_variant = parsed_string.root_variant()
            if not root_variant:
                # TODO: Make this prettier
                output_stream.write(
                    self.colorize("...Failed to Parse...", Color.red)
                )  # pragma: no cover
            elif num_variants == 1:
                # Backward compatible single parse
                assert root_variant.tree
                output_stream.write(root_variant.tree.stringify(code_only=code_only))
            else:
                # Multi variant parse setup.
                output_stream.write(
                    self.colorize(
                        f"SQLFluff parsed {num_variants} variants of this file",
                        Color.blue,
                    )
                )
                for idx, variant in enumerate(parsed_string.parsed_variants):
                    output_stream.write(
                        self.colorize(
                            f"Variant {idx + 1}:",
                            Color.blue,
                        )
                    )
                    if variant.tree:
                        output_stream.write(variant.tree.stringify(code_only=code_only))
                    else:  # pragma: no cover
                        output_stream.write(
                            self.colorize("...Failed to Parse...", Color.red)
                        )

            violations = (
                violations_getter(parsed_string)
                if violations_getter
                else parsed_string.violations
            )
            violations_count += len(violations)
            if violations:
                output_stream.write("==== parsing violations ====")  # pragma: no cover
            for v in violations:
                output_stream.write(self.format_violation(v))  # pragma: no cover
            if violations:
                output_stream.write(
                    self.format_dialect_warning(parsed_string.config.get("dialect"))
                )

            if self.output_policy.allows(OutputKind.VERBOSE, minimum_verbosity=2):
                output_stream.write("==== timings ====")
                output_stream.write(self.cli_table(parsed_string.time_dict.items()))

        if self.output_policy.allows(OutputKind.VERBOSE, minimum_verbosity=2) or bench:
            output_stream.write("==== overall timings ====")
            output_stream.write(self.cli_table([("Clock time", total_time)]))
            timing_summary = timing.summary()
            for step in timing_summary:
                output_stream.write(f"=== {step} ===")
                output_stream.write(self.cli_table(timing_summary[step].items()))

        return violations_count

    def completion_message(self) -> None:
        """Prints message when SQLFluff is finished."""
        self.dispatch_message(
            f"All Finished{'' if self.plain_output else ' 📜 🎉'}!",
            OutputKind.STATUS,
        )



"""A list of db2 keywords."""

UNRESERVED_KEYWORDS = [
    # https://www.ibm.com/docs/en/db2/11.5?topic=sql-reserved-schema-names-reserved-words
    "ACTIVATE",
    "ADD",
    "AFTER",
    "AGE",
    "ALIAS",
    "ALL",
    "ALLOCATE",
    "ALLOW",
    "ALTER",
    "AND",
    "ANY",
    "AS",
    "ASENSITIVE",
    "ASSOCIATE",
    "ASUTIME",
    "ASYNCHRONY",
    "AT",
    "ATTRIBUTES",
    "AUDIT",
    "AUTHORIZATION",
    "AUX",
    "AUXILIARY",
    "BEFORE",
    "BEGIN",
    "BETWEEN",
    "BINARY",
    "BUFFERPOOL",
    "BUSINESS_TIME",
    "BY",
    "CACHE",
    "CALL",
    "CALLED",
    "CAPTURE",
    "CARDINALITY",
    "CASCADED",
    "CASE",
    "CAST",
    "CCSID",
    "CHAR",
    "CHARACTER",
    "CHECK",
    "CLIENT_ACCTNG",
    "CLIENT_APPLNAME",
    "CLIENT_USERID",
    "CLIENT_WRKSTNNAME",
    "CLONE",
    "CLOSE",
    "CLUSTER",
    "COLLECTION",
    "COLLID",
    "COLUMN",
    "COMMENT",
    "COMMIT",
    "CONCAT",
    "CONDITION",
    "CONNECT",
    "CONNECTION",
    "CONSTRAINT",
    "CONTAINS",
    "CONTINUE",
    "COUNT",
    "COUNT_BIG",
    "CREATE",
    "CROSS",
    "CS",
    "CURRENT",
    "CURRENT_DATE",
    "CURRENT_LC_CTYPE",
    "CURRENT_PATH",
    "CURRENT_SCHEMA",
    "CURRENT_SERVER",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_TIMEZONE",
    "CURRENT_USER",
    "CURSOR",
    "CYCLE",
    "DATA",
    "DATABASE",
    "DATAPARTITIONNAME",
    "DATAPARTITIONNUM",
    "DATE",
    "DAY",
    "DAYS",
    "DB2GENERAL",
    "DB2GENRL",
    "DB2SQL",
    "DBINFO",
    "DBPARTITIONNAME",
    "DBPARTITIONNUM",
    "DEALLOCATE",
    "DECFLOAT",
    "DECLARE",
    "DEFAULT",
    "DEFAULTS",
    "DEFINITION",
    "DELETE",
    "DENSERANK",
    "DENSE_RANK",
    "DESCRIBE",
    "DESCRIPTOR",
    "DETAILED",
    "DETERMINISTIC",
    "DIAGNOSTICS",
    "DISABLE",
    "DISALLOW",
    "DISCONNECT",
    "DISTINCT",
    "DISTRIBUTE",
    "DO",
    "DOCUMENT",
    "DOUBLE",
    "DROP",
    "DSSIZE",
    "DYNAMIC",
    "EACH",
    "EDITPROC",
    "ELSE",
    "ELSEIF",
    "ENABLE",
    "ENCODING",
    "ENCRYPTION",
    "END",
    "END-EXEC",
    "ENDING",
    "ERASE",
    "ESCAPE",
    "EVERY",
    "EXCEPT",
    "EXCEPTION",
    "EXCLUDING",
    "EXCLUSIVE",
    "EXECUTE",
    "EXISTS",
    "EXIT",
    "EXPLAIN",
    "EXTEND",
    "EXTENDED",
    "EXTERNAL",
    "EXTRACT",
    "FEDERATED",
    "FENCED",
    "FETCH",
    "FIELDPROC",
    "FILE",
    "FINAL",
    "FIRST",
    "FOR",
    "FOREIGN",
    "FREE",
    "FROM",
    "FULL",
    "FUNCTION",
    "GENERAL",
    "GENERATE",
    "GENERATED",
    "GET",
    "GLOBAL",
    "GO",
    "GOTO",
    "GRANT",
    "GRAPHIC",
    "GROUP",
    "HANDLER",
    "HASH",
    "HASHED_VALUE",
    "HAVING",
    "HIGH",
    "HINT",
    "HOLD",
    "HOUR",
    "HOURS",
    "IDENTITY",
    "IF",
    "IMMEDIATE",
    "IMPORT",
    "IN",
    "INCLUDING",
    "INCLUSIVE",
    "INCREMENT",
    "INDEX",
    "INDICATOR",
    "INDICATORS",
    "INF",
    "INFINITY",
    "INHERIT",
    "INNER",
    "INOUT",
    "INSENSITIVE",
    "INSERT",
    "INTEGRITY",
    "INTERSECT",
    "INTO",
    "IS",
    "ISNULL",
    "ISOBID",
    "ISOLATION",
    "ITERATE",
    "JAR",
    "JAVA",
    "JOIN",
    "KEEP",
    "KEY",
    "LABEL",
    "LANGUAGE",
    "LAST",
    "LATERAL",
    "LC_CTYPE",
    "LC_MESSAGES",
    "LC_TIME",
    "LEAVE",
    "LEFT",
    "LEVEL2",
    "LIKE",
    "LIMIT",
    "LINKTYPE",
    "LOCAL",
    "LOCALDATE",
    "LOCALE",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "LOCATOR",
    "LOCATORS",
    "LOCK",
    "LOCKMAX",
    "LOCKSIZE",
    "LOGGED",
    "LONG",
    "LOOP",
    "LOW",
    "MAINTAINED",
    "MATERIALIZED",
    "MAXVALUE",
    "MDC",
    "MICROSECOND",
    "MICROSECONDS",
    "MINPCTUSED",
    "MINUS",
    "MINUTE",
    "MINUTES",
    "MINVALUE",
    "MODE",
    "MODIFIES",
    "MONTH",
    "MONTHS",
    "MOVEMENT",
    "NAN",
    "NEW",
    "NEW_TABLE",
    "NEXTVAL",
    "NO",
    "NOCACHE",
    "NOCYCLE",
    "NODENAME",
    "NODENUMBER",
    "NOMAXVALUE",
    "NOMINVALUE",
    "NONE",
    "NOORDER",
    "NORMALIZED",
    "NOT",
    "NOTNULL",
    "NULL",
    "NULLS",
    "NUMPARTS",
    "OBID",
    "OF",
    "OFF",
    "OFFSET",
    "OLD",
    "OLD_TABLE",
    "ON",
    "OPEN",
    "OPTIMIZATION",
    "OPTIMIZE",
    "OPTION",
    "OR",
    "ORDER",
    "ORGANIZE",
    "OUT",
    "OUTER",
    "OVER",
    "OVERRIDING",
    "PACKAGE",
    "PADDED",
    "PAGE",
    "PAGESIZE",
    "PARAMETER",
    "PART",
    "PARTITION",
    "PARTITIONED",
    "PARTITIONING",
    "PARTITIONS",
    "PASSWORD",
    "PATH",
    "PERCENT",
    "PIECESIZE",
    "PLAN",
    "POSITION",
    "PRECISION",
    "PREPARE",
    "PREVVAL",
    "PRIMARY",
    "PRIQTY",
    "PRIVILEGES",
    "PROCEDURE",
    "PROFILE",
    "PROGRAM",
    "PSID",
    "PUBLIC",
    "QUERY",
    "QUERYNO",
    "RANDOM",
    "RANGE",
    "RANK",
    "READ",
    "READS",
    "RECOVERY",
    "REFERENCES",
    "REFERENCING",
    "REFRESH",
    "REJECT",
    "RELEASE",
    "RENAME",
    "REPEAT",
    "RESET",
    "RESIGNAL",
    "RESTART",
    "RESTRICT",
    "RESULT",
    "RESULT_SET_LOCATOR",
    "RETURN",
    "RETURNS",
    "REVERSE",
    "REVOKE",
    "RIGHT",
    "ROLE",
    "ROLLBACK",
    "ROLLOUT",
    "ROUND_CEILING",
    "ROUND_DOWN",
    "ROUND_FLOOR",
    "ROUND_HALF_DOWN",
    "ROUND_HALF_EVEN",
    "ROUND_HALF_UP",
    "ROUND_UP",
    "ROUNDING",
    "ROUTINE",
    "ROW",
    "ROWNUMBER",
    "ROWS",
    "ROWSET",
    "ROW_NUMBER",
    "RR",
    "RRN",
    "RS",
    "RUN",
    "SAVEPOINT",
    "SAMPLED",
    "SCANS",
    "SCHEMA",
    "SCRATCHPAD",
    "SCROLL",
    "SEARCH",
    "SECOND",
    "SECONDS",
    "SECQTY",
    "SECURITY",
    "SELECT",
    "SENSITIVE",
    "SEQUENCE",
    "SESSION",
    "SESSION_USER",
    "SET",
    "SIGNAL",
    "SIMPLE",
    "SNAN",
    "SNAPSHOT",
    "SOME",
    "SOURCE",
    "SPECIFIC",
    "SPECIFICATION",
    "SPLIT",
    "SQL",
    "SQL_CCFLAGS",
    "SQLID",
    "STACKED",
    "STANDARD",
    "START",
    "STARTING",
    "STATEMENT",
    "STATIC",
    "STATMENT",
    "STAY",
    "STOGROUP",
    "STORES",
    "STYLE",
    "SUBSTRING",
    "SUMMARY",
    "SYNONYM",
    "SYSFUN",
    "SYSIBM",
    "SYSPROC",
    "SYSTEM",
    "SYSTEM_TIME",
    "SYSTEM_USER",
    "TABLE",
    "TABLESPACE",
    "TEMPORAL",
    "THEN",
    "TIME",
    "TIMEOUT",
    "TIMESTAMP",
    "TIMEZONE",
    "TO",
    "TRANSACTION",
    "TRIGGER",
    "TRIM",
    "TRUNCATE",
    "TYPE",
    "TYPES",
    "UNDO",
    "UNION",
    "UNIQUE",
    "UNTIL",
    "UNSAMPLED",
    "UPDATE",
    "UR",
    "USAGE",
    "USER",
    "USING",
    "VALIDPROC",
    "VALUE",
    "VALUES",
    "VARIABLE",
    "VARIANT",
    "VCAT",
    "VERSION",
    "VIEW",
    "VOLATILE",
    "VOLUMES",
    "WHEN",
    "WHENEVER",
    "WHERE",
    "WHILE",
    "WITH",
    "WITHOUT",
    "WLM",
    "WRITE",
    "XMLELEMENT",
    "XMLEXISTS",
    "XMLNAMESPACES",
    "XMLPARSE",
    "XMLPATTERN",
    "YEAR",
    "YEARS",
    "YES",
]



"""The convention plugin bundle."""

from sqlfluff.core.plugin import hookimpl
from sqlfluff.core.rules import BaseRule, ConfigInfo


@hookimpl
def get_configs_info() -> dict[str, ConfigInfo]:
    """Get additional rule config validations and descriptions."""
    return {
        "preferred_not_equal_style": {
            "validation": ["consistent", "c_style", "ansi"],
            "definition": (
                "The style for using not equal to operator. Defaults to ``consistent``."
            ),
        },
        "select_clause_trailing_comma": {
            "validation": ["forbid", "require"],
            "definition": (
                "Should trailing commas within select clauses be required or forbidden?"
            ),
        },
        "prefer_count_1": {
            "validation": [True, False],
            "definition": ("Should count(1) be preferred over count(*) and count(0)?"),
        },
        "prefer_count_0": {
            "validation": [True, False],
            "definition": ("Should count(0) be preferred over count(*) and count(1)?"),
        },
        "multiline_newline": {
            "validation": [True, False],
            "definition": (
                "Should semi-colons be placed on a new line after multi-line "
                "statements?"
            ),
        },
        "require_final_semicolon": {
            "validation": [True, False],
            "definition": (
                "Should final semi-colons be required? "
                "(N.B. forcing trailing semi-colons is not recommended for dbt users "
                "as it can cause issues when wrapping the query within other SQL "
                "queries)."
            ),
        },
        "preferred_quoted_literal_style": {
            "validation": ["consistent", "single_quotes", "double_quotes"],
            "definition": (
                "Preferred quoting style to use for the quoted literals. If set to "
                "``consistent`` quoting style is derived from the first quoted literal "
                "in the file."
            ),
        },
        "preferred_type_casting_style": {
            "validation": ["consistent", "shorthand", "convert", "cast"],
            "definition": ("The expectation for using sql type casting"),
        },
    }


@hookimpl
def get_rules() -> list[type[BaseRule]]:
    """Get plugin rules.

    NOTE: Rules are imported only on fetch to manage import times
    when rules aren't used.
    """
    from sqlfluff.rules.convention.CV01 import Rule_CV01
    from sqlfluff.rules.convention.CV02 import Rule_CV02
    from sqlfluff.rules.convention.CV03 import Rule_CV03
    from sqlfluff.rules.convention.CV04 import Rule_CV04
    from sqlfluff.rules.convention.CV05 import Rule_CV05
    from sqlfluff.rules.convention.CV06 import Rule_CV06
    from sqlfluff.rules.convention.CV07 import Rule_CV07
    from sqlfluff.rules.convention.CV08 import Rule_CV08
    from sqlfluff.rules.convention.CV09 import Rule_CV09
    from sqlfluff.rules.convention.CV10 import Rule_CV10
    from sqlfluff.rules.convention.CV11 import Rule_CV11
    from sqlfluff.rules.convention.CV12 import Rule_CV12

    return [
        Rule_CV01,
        Rule_CV02,
        Rule_CV03,
        Rule_CV04,
        Rule_CV05,
        Rule_CV06,
        Rule_CV07,
        Rule_CV08,
        Rule_CV09,
        Rule_CV10,
        Rule_CV11,
        Rule_CV12,
    ]



"""The DuckDB dialect.

https://duckdb.org/docs/
"""

from sqlfluff.core.dialects import load_raw_dialect
from sqlfluff.core.parser import (
    AnyNumberOf,
    BaseSegment,
    BinaryOperatorSegment,
    Bracketed,
    CodeSegment,
    ComparisonOperatorSegment,
    Dedent,
    Delimited,
    IdentifierSegment,
    ImplicitIndent,
    Indent,
    Matchable,
    Nothing,
    OneOf,
    OptionallyBracketed,
    ParseMode,
    Ref,
    RegexLexer,
    RegexParser,
    Sequence,
    StringLexer,
    StringParser,
    SymbolSegment,
    TypedParser,
)
from sqlfluff.dialects import dialect_ansi as ansi
from sqlfluff.dialects import dialect_postgres as postgres

ansi_dialect = load_raw_dialect("ansi")
postgres_dialect = load_raw_dialect("postgres")
duckdb_dialect = postgres_dialect.copy_as(
    "duckdb",
    formatted_name="DuckDB",
    docstring="""**Default Casing**: DuckDB stores all identifiers in the case
they were defined, however all identifier resolution is case-insensitive (when
unquoted, and more unusually, *also when quoted*). See the
`DuckDB Identifiers Documentation`_ for more details.

**Quotes**: String Literals: ``''``, Identifiers: ``""`` or ``''``

The dialect for `DuckDB <https://duckdb.org/>`_.

.. _`DuckDB Identifiers Documentation`: https://duckdb.org/docs/sql/dialect/keywords_and_identifiers
""",
)

duckdb_dialect.sets("reserved_keywords").update(
    [
        "LAMBDA",
        "PIVOT",
        "PIVOT_LONGER",
        "PIVOT_WIDER",
        "UNPIVOT",
    ]
)

duckdb_dialect.sets("unreserved_keywords").update(
    [
        "ANTI",
        "APPEND",
        "ASOF",
        "COMPRESSION",
        "COMPRESSION_LEVEL",
        "GLOB",
        "MACRO",
        "MAP",
        "OVERWRITE",
        "OVERWRITE_OR_IGNORE",
        "PARQUET_VERSION",
        "PARTITION_BY",
        "PERCENT",
        "POSITIONAL",
        "PROGRAM",
        "RESERVOIR",
        "ROW_GROUP_SIZE",
        "ROW_GROUP_SIZE_BYTES",
        "SAMPLE",
        "SEMI",
        "STRUCT",
        "VIRTUAL",
        "WRITE_PARTITION_COLUMNS",
    ]
)

# Add plural datetime units for interval expressions
# DuckDB supports both singular and plural forms (e.g., DAY and DAYS)
# Note: ANSI has MILLISECOND (singular), PostgreSQL adds MILLISECONDS and
# MICROSECONDS (both plural). We add the missing MICROSECOND (singular) and
# all other plural forms (DAYS, HOURS, MINUTES, MONTHS, QUARTERS, SECONDS,
# WEEKS, YEARS).
duckdb_dialect.sets("datetime_units").update(
    [
        "DAYS",
        "HOURS",
        "MICROSECOND",  # Singular form missing from ANSI/PostgreSQL
        "MINUTES",
        "MONTHS",
        "QUARTERS",
        "SECONDS",
        "WEEKS",
        "YEARS",
    ]
)

duckdb_dialect.add(
    LambdaArrowSegment=StringParser("->", SymbolSegment, type="lambda_arrow"),
    OrIgnoreGrammar=Sequence("OR", "IGNORE"),
    # DuckDB sampling. A percentage (`10%` / `10 PERCENT`) as opposed to a fixed
    # row count (a bare number, or `10 ROWS`). Only reservoir sampling supports a
    # fixed row count; bernoulli and system sampling are percentage-only. Shared
    # by the table-level `TABLESAMPLE` and the query-level `USING SAMPLE` clauses.
    # https://duckdb.org/docs/stable/sql/samples
    SamplePercentageGrammar=Sequence(
        Ref("NumericLiteralSegment"), OneOf(Ref("ModuloSegment"), "PERCENT")
    ),
    SampleRowCountGrammar=Sequence(
        Ref("NumericLiteralSegment"), Ref.keyword("ROWS", optional=True)
    ),
    SampleExpressionGrammar=Sequence(
        OneOf(
            # reservoir takes a percentage or a fixed row count
            Sequence(
                "RESERVOIR",
                Bracketed(
                    OneOf(Ref("SamplePercentageGrammar"), Ref("SampleRowCountGrammar"))
                ),
            ),
            # bernoulli and system are percentage-only
            Sequence(
                OneOf("BERNOULLI", "SYSTEM"),
                Bracketed(Ref("SamplePercentageGrammar")),
            ),
            # a percentage, optionally with any method (and seed)
            Sequence(
                Ref("SamplePercentageGrammar"),
                Bracketed(
                    OneOf("RESERVOIR", "BERNOULLI", "SYSTEM"),
                    Sequence(
                        Ref("CommaSegment"),
                        Ref("NumericLiteralSegment"),
                        optional=True,
                    ),
                    optional=True,
                ),
            ),
            # a fixed row count, optionally with reservoir (and seed)
            Sequence(
                Ref("SampleRowCountGrammar"),
                Bracketed(
                    "RESERVOIR",
                    Sequence(
                        Ref("CommaSegment"),
                        Ref("NumericLiteralSegment"),
                        optional=True,
                    ),
                    optional=True,
                ),
            ),
        ),
        Sequence("REPEATABLE", Bracketed(Ref("NumericLiteralSegment")), optional=True),
    ),
    EqualsSegment_a=StringParser("==", ComparisonOperatorSegment),
    UnpackingOperatorSegment=TypedParser("star", SymbolSegment, "unpacking_operator"),
    # DuckDB math operators
    PowerOperatorSegment=TypedParser(
        "power_operator", SymbolSegment, type="binary_operator"
    ),
    AbsoluteValueOperatorSegment=TypedParser(
        "at", SymbolSegment, type="sign_indicator"
    ),
    FactorialOperatorSegment=TypedParser(
        "not", SymbolSegment, type="factorial_operator"
    ),
)

duckdb_dialect.replace(
    FunctionNameIdentifierSegment=RegexParser(
        r"[A-Z_][A-Z0-9_$]*",
        CodeSegment,
        type="function_name_identifier",
        anti_template=r"^(STRUCT|UNION|ENUM)$",
    ),
    DivideSegment=OneOf(
        StringParser("//", BinaryOperatorSegment),
        StringParser("/", BinaryOperatorSegment),
    ),
    CreateTableAsStatementSegment=Nothing(),
    UnionGrammar=ansi_dialect.get_grammar("UnionGrammar").copy(
        insert=[
            Sequence("BY", "NAME", optional=True),
        ]
    ),
    JoinLikeClauseGrammar=Sequence(
        AnyNumberOf(
            Ref("FromPivotExpressionSegment"),
            Ref("FromUnpivotExpressionSegment"),
            min_times=1,
        ),
        Ref("AliasExpressionSegment", optional=True),
    ),
    NonSetSelectableGrammar=postgres_dialect.get_grammar(
        "NonSetSelectableGrammar"
    ).copy(
        insert=[
            Ref("SimplifiedPivotExpressionSegment"),
            Ref("SimplifiedUnpivotExpressionSegment"),
        ],
    ),
    NonStandardJoinTypeKeywordsGrammar=OneOf(
        "ANTI",
        "SEMI",
        Sequence(
            "ASOF",
            OneOf(
                Ref("JoinTypeKeywordsGrammar"),
                "ANTI",
                "SEMI",
                optional=True,
            ),
        ),
    ),
    HorizontalJoinKeywordsGrammar=Ref.keyword("POSITIONAL"),
    FunctionContentsExpressionGrammar=OneOf(
        Ref("LambdaExpressionSegment"),
        Ref("NamedArgumentSegment"),
        Ref("ExpressionSegment"),
    ),
    ColumnsExpressionNameGrammar=Sequence(
        Ref("UnpackingOperatorSegment", optional=True), "COLUMNS"
    ),
    # Uses grammar for LT06 support
    ColumnsExpressionGrammar=Sequence(
        Ref("ColumnsExpressionFunctionNameSegment"),
        Ref("ColumnsExpressionFunctionContentsSegment"),
    ),
    # Matching postgres lower casefold, as it is case-insensitive
    QuotedIdentifierSegment=TypedParser(
        "double_quote", IdentifierSegment, type="quoted_identifier", casefold=str.lower
    ),
    SingleQuotedIdentifierSegment=TypedParser(
        "single_quote", IdentifierSegment, type="quoted_identifier", casefold=str.lower
    ),
    ListComprehensionGrammar=Ref("ListComprehensionExpressionSegment"),
    # non-ANSI IN operator defined for string, list, and map
    # https://duckdb.org/docs/stable/sql/expressions/in
    InOperatorGrammar=Sequence(
        Ref.keyword("NOT", optional=True),
        "IN",
        OneOf(
            Bracketed(
                OneOf(
                    Delimited(Ref("Expression_A_Grammar"), allow_trailing=True),
                    Ref("SelectableGrammar"),
                ),
                parse_mode=ParseMode.GREEDY,
            ),
            Ref("FunctionSegment"),
            Ref("ArrayLiteralSegment"),
            Ref("MapLiteralSegment"),
            Ref("QuotedLiteralSegment"),
            Ref("ColumnReferenceSegment"),
        ),
    ),
    # Add DuckDB math operators
    ArithmeticBinaryOperatorGrammar=postgres_dialect.get_grammar(
        "ArithmeticBinaryOperatorGrammar"
    ).copy(
        insert=[
            Ref("PowerOperatorSegment"),
        ]
    ),
    # Add @ prefix operator for absolute value
    Expression_A_Unary_Operator_Grammar=postgres_dialect.get_grammar(
        "Expression_A_Unary_Operator_Grammar"
    ).copy(
        insert=[
            Ref("AbsoluteValueOperatorSegment"),
        ]
    ),
    # Add postfix factorial operator support
    Expression_C_Grammar=Sequence(
        postgres_dialect.get_grammar("Expression_C_Grammar"),
        Ref(
            "FactorialOperatorSegment", exclude=Ref("NotEqualToSegment"), optional=True
        ),
    ),
    ComparisonOperatorGrammar=ansi_dialect.get_grammar(
        "ComparisonOperatorGrammar"
    ).copy(
        insert=[
            Ref("EqualsSegment_a"),
            Ref("GlobOperatorSegment"),
        ]
    ),
    LikeGrammar=postgres_dialect.get_grammar("LikeGrammar").copy(
        insert=[Ref.keyword("GLOB")],
    ),
    FilterClauseGrammar=Sequence(
        "FILTER",
        Bracketed(
            Sequence(Ref.keyword("WHERE", optional=True), Ref("ExpressionSegment"))
        ),
    ),
    BaseExpressionElementGrammar=OneOf(
        Ref("MapLiteralSegment"),
        postgres_dialect.get_grammar("BaseExpressionElementGrammar"),
        Ref("ColumnIndexSegment"),
    ),
)

# Patch lexers before adding segments
duckdb_dialect.patch_lexer_matchers(
    [
        # Remove @ from postgis_operator regex since
        # we use it for absolute value
        # Maybe even the postgres version is strange,
        # as postgis operators shouldn't use @ either?
        RegexLexer(
            "postgis_operator",
            r"\&\&\&|\&<\||<<\||\|\&>|\|>>|\~=|<\->|\|=\||<\#>|<<\->>|<<\#>>",
            SymbolSegment,
        ),
    ]
)

duckdb_dialect.insert_lexer_matchers(
    [
        StringLexer("power_operator", "**", CodeSegment),
    ],
    before="star",
)

duckdb_dialect.insert_lexer_matchers(
    [
        StringLexer("double_divide", "//", CodeSegment),
        # Column indexes: #1, #2, etc.
        RegexLexer("column_index", r"#[0-9]+", CodeSegment),
    ],
    before="divide",
)

duckdb_dialect.patch_lexer_matchers(
    [
        # In DuckDB, a double single/double quote resolves as a single/double quote in
        # the string.
        RegexLexer(
            "single_quote",
            r"'([^']|'')*'",
            CodeSegment,
            segment_kwargs={
                "quoted_value": (r"'((?:[^']|'')*)'", 1),
                "escape_replacements": [(r"''", "'")],
            },
        ),
        RegexLexer(
            "double_quote",
            r'"([^"]|"")*"',
            CodeSegment,
            segment_kwargs={
                "quoted_value": (r'"((?:[^"]|"")*)"', 1),
                "escape_replacements": [(r'""', '"')],
            },
        ),
        RegexLexer("equals", r"==?", CodeSegment),
    ]
)


class SamplingExpressionSegment(ansi.SamplingExpressionSegment):
    """A DuckDB ``TABLESAMPLE`` expression.

    ``TABLESAMPLE`` is the table-level form: it samples an individual table in
    the ``FROM`` clause (before joins). The query-level ``USING SAMPLE`` form is
    handled by ``UsingSampleClauseSegment``.

    https://duckdb.org/docs/stable/sql/samples
    """

    match_grammar: Matchable = Sequence(
        "TABLESAMPLE",
        Ref("SampleExpressionGrammar"),
    )


class UsingSampleClauseSegment(BaseSegment):
    """A DuckDB query-level ``USING SAMPLE`` clause.

    Unlike ``TABLESAMPLE``, ``USING SAMPLE`` samples the result of the whole
    ``FROM`` clause (after joins), so it trails the query rather than attaching
    to a single table.

    https://duckdb.org/docs/stable/sql/samples
    """

    type = "sample_expression"
    match_grammar: Matchable = Sequence(
        "USING",
        "SAMPLE",
        Ref("SampleExpressionGrammar"),
    )


class IntervalExpressionSegment(BaseSegment):
    """An interval expression segment.

    Extends ANSI to support dynamic intervals with parenthesized expressions.
    https://duckdb.org/docs/stable/sql/data_types/interval
    """

    type = "interval_expression"
    match_grammar: Matchable = Sequence(
        "INTERVAL",
        OneOf(
            # The Numeric Version
            Sequence(
                Ref("NumericLiteralSegment"),
                OneOf(Ref("QuotedLiteralSegment"), Ref("DatetimeUnitSegment")),
            ),
            # The String version
            Ref("QuotedLiteralSegment"),
            # Combine version
            Sequence(
                Ref("QuotedLiteralSegment"),
                OneOf(Ref("QuotedLiteralSegment"), Ref("DatetimeUnitSegment")),
            ),
            # Expression version - for dynamic intervals with parenthesized expressions
            Sequence(
                Ref("ExpressionSegment"),
                Ref("DatetimeUnitSegment"),
            ),
        ),
    )


class ColumnIndexSegment(BaseSegment):
    """Column index reference using positional syntax.

    DuckDB allows referencing columns by their position using #<number> syntax.
    https://duckdb.org/docs/stable/sql/statements/select
    """

    type = "column_index"
    match_grammar = TypedParser(
        "column_index",
        CodeSegment,
        type="column_index",
    )


class StructTypeSegment(ansi.StructTypeSegment):
    """Expression to construct a STRUCT datatype."""

    match_grammar = Sequence(
        "STRUCT",
        Ref("StructTypeSchemaSegment", optional=True),
    )


class StructTypeSchemaSegment(BaseSegment):
    """Expression to construct the schema of a STRUCT datatype."""

    type = "struct_type_schema"
    match_grammar = Bracketed(
        Delimited(  # Comma-separated list of field names/types
            Sequence(
                OneOf(
                    # ParameterNames can look like Datatypes so can't use
                    # Optional=True here and instead do a OneOf in order
                    # with DataType only first, followed by both.
                    Ref("DatatypeSegment"),
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref("DatatypeSegment"),
                    ),
                ),
            ),
        ),
    )


class MapTypeSegment(ansi.MapTypeSegment):
    """Expression to construct a MAP datatype."""

    match_grammar = Sequence(
        "MAP",
        Ref("MapTypeSchemaSegment", optional=True),
    )


class MapTypeSchemaSegment(BaseSegment):
    """Expression to construct the schema of a MAP datatype."""

    type = "map_type_schema"
    match_grammar = Bracketed(
        Delimited(
            Ref("DatatypeSegment"),
        ),
    )


class MapLiteralElementSegment(BaseSegment):
    """A map literal element segment.

    e.g. 'key1': 50
    Keys in DuckDB map literals can be any expression.
    """

    type = "map_literal_element"
    match_grammar: Matchable = Sequence(
        Ref("BaseExpressionElementGrammar"),
        Ref("ColonSegment"),
        Ref("BaseExpressionElementGrammar"),
    )


class MapLiteralSegment(BaseSegment):
    """A map literal segment.

    https://duckdb.org/docs/stable/sql/data_types/map
    e.g. MAP {'key1': 50, 'key2': 75}
    """

    type = "map_literal"
    match_grammar: Matchable = Sequence(
        "MAP",
        Bracketed(
            Delimited(
                Ref("MapLiteralElementSegment"),
                optional=True,
                allow_trailing=True,
            ),
            bracket_type="curly",
        ),
    )


class InsertStatementSegment(ansi.InsertStatementSegment):
    """An `INSERT` Statement.

    https://duckdb.org/docs/stable/sql/statements/insert.html
    """

    type = "insert_statement"
    match_grammar: Matchable = Sequence(
        "INSERT",
        OneOf(Ref("OrReplaceGrammar"), Ref("OrIgnoreGrammar"), optional=True),
        "INTO",
        Ref("TableReferenceSegment"),
        Ref("AsAliasExpressionSegment", optional=True),
        OneOf(
            Ref("BracketedColumnReferenceListGrammar"),
            Sequence("BY", "POSITION"),
            Sequence("BY", "NAME"),
            optional=True,
        ),
        OneOf(
            Sequence("DEFAULT", "VALUES"),
            Ref("SelectableGrammar"),
            Sequence(
                Ref("BracketedColumnReferenceListGrammar", optional=True),
                OneOf(
                    Ref("ValuesClauseSegment"),
                    OptionallyBracketed(Ref("SelectableGrammar")),
                ),
            ),
        ),
        Sequence(
            "ON",
            "CONFLICT",
            Ref("ConflictTargetSegment", optional=True),
            Ref("ConflictActionSegment"),
            optional=True,
        ),
        Sequence(
            "RETURNING",
            OneOf(
                Ref("StarSegment"),
                Delimited(
                    Sequence(
                        Ref("ExpressionSegment"),
                        Ref("AsAliasExpressionSegment", optional=True),
                    ),
                ),
            ),
            optional=True,
        ),
    )


class ColumnConstraintSegment(ansi.ColumnConstraintSegment):
    """A column option; each CREATE TABLE column can have 0 or more.

    https://duckdb.org/docs/sql/statements/create_table
    https://duckdb.org/docs/sql/statements/alter_table
    """

    # Column constraint from
    # https://duckdb.org/docs/sql/statements/create_table
    match_grammar = Sequence(
        OneOf(
            Sequence(Ref.keyword("NOT", optional=True), "NULL"),  # NOT NULL or NULL
            Sequence(
                "CHECK",
                Bracketed(Ref("ExpressionSegment")),
            ),
            Sequence(  # DEFAULT <value>
                "DEFAULT",
                OneOf(
                    Ref("LiteralGrammar"),
                    Ref("ExpressionSegment"),
                ),
            ),
            "UNIQUE",
            Sequence(
                "PRIMARY",
                "KEY",
            ),
            Ref("ReferenceDefinitionGrammar"),
            Sequence(
                "COLLATE",
                Ref("CollationReferenceSegment"),
            ),
        ),
    )


class CreateTableStatementSegment(ansi.CreateTableStatementSegment):
    """A `CREATE TABLE` statement.

    As specified in https://duckdb.org/docs/sql/statements/create_table.html
    """

    match_grammar = Sequence(
        "CREATE",
        Ref("OrReplaceGrammar", optional=True),
        Ref("TemporaryGrammar", optional=True),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        OneOf(
            Sequence(
                "AS",
                OptionallyBracketed(Ref("SelectableGrammar")),
            ),
            # Columns and comment syntax:
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("ColumnReferenceSegment"),
                        OneOf(
                            Sequence(
                                Ref("DatatypeSegment"),
                                AnyNumberOf(
                                    OneOf(
                                        Ref("ColumnConstraintSegment"),
                                    ),
                                ),
                            ),
                            Sequence(
                                Ref(
                                    "DatatypeSegment",
                                    optional=True,
                                    exclude=Ref.keyword("AS"),
                                ),
                                Sequence("GENERATED", "ALWAYS", optional=True),
                                "AS",
                                Bracketed(Ref("ExpressionSegment")),
                                OneOf("STORED", "VIRTUAL", optional=True),
                            ),
                        ),
                    ),
                    Ref("TableConstraintSegment"),
                )
            ),
        ),
    )


class WildcardExcludeExpressionSegment(BaseSegment):
    """An `EXCLUDE` clause within a wildcard expression."""

    type = "wildcard_exclude"
    match_grammar = Sequence(
        "EXCLUDE",
        OneOf(
            Ref("ColumnReferenceSegment"),
            Bracketed(Delimited(Ref("ColumnReferenceSegment"), allow_trailing=True)),
        ),
    )


class WildcardReplaceExpressionSegment(BaseSegment):
    """A `REPLACE` clause within a wildcard expression."""

    type = "wildcard_replace"
    match_grammar = Sequence(
        "REPLACE",
        OneOf(
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("BaseExpressionElementGrammar"),
                        Ref("AliasExpressionSegment", optional=True),
                    ),
                    allow_trailing=True,
                )
            ),
            Sequence(
                Ref("BaseExpressionElementGrammar"),
                Ref("AliasExpressionSegment", optional=True),
            ),
        ),
    )


class WildcardRenameExpressionSegment(BaseSegment):
    """A `RENAME` clause within a wildcard expression."""

    type = "wildcard_rename"
    match_grammar = Sequence(
        "RENAME",
        OneOf(
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("BaseExpressionElementGrammar"),
                        Ref("AliasExpressionSegment", optional=True),
                    ),
                    allow_trailing=True,
                )
            ),
            Sequence(
                Ref("BaseExpressionElementGrammar"),
                Ref("AliasExpressionSegment", optional=True),
            ),
        ),
    )


class WildcardPatternMatchingSegment(BaseSegment):
    """A pattern matching operator clause within a wildcard expression."""

    type = "wildcard_pattern_matching"
    match_grammar = OneOf(
        Ref("LikeExpressionGrammar"),
        Sequence(
            OneOf(
                Ref("LikeOperatorSegment"),
                Ref("GlobOperatorSegment"),
            ),
            Ref("QuotedLiteralSegment"),
        ),
    )


class WildcardExpressionSegment(ansi.WildcardExpressionSegment):
    """An extension of the star expression for DuckDB."""

    match_grammar = Sequence(
        # *, blah.*, blah.blah.*, etc.
        Ref("WildcardIdentifierSegment"),
        # Optional EXCLUDE or REPLACE clause
        Ref("WildcardExcludeExpressionSegment", optional=True),
        OneOf(
            Sequence(
                Ref("WildcardReplaceExpressionSegment", optional=True),
                Ref("WildcardRenameExpressionSegment", optional=True),
            ),
            Ref("WildcardPatternMatchingSegment"),
            optional=True,
        ),
    )


class SelectClauseElementSegment(ansi.SelectClauseElementSegment):
    """An element in the targets of a select statement."""

    type = "select_clause_element"

    match_grammar = OneOf(
        Sequence(
            Ref("WildcardExpressionSegment"),
        ),
        Sequence(
            Ref(
                "BaseExpressionElementGrammar",
            ),
            Ref("AliasExpressionSegment", optional=True),
        ),
    )


class ColumnsExpressionFunctionContentsSegment(
    ansi.ColumnsExpressionFunctionContentsSegment
):
    """Columns expression in a select statement.

    https://duckdb.org/docs/sql/expressions/star#columns-expression
    """

    type = "function_contents"
    match_grammar = Sequence(
        Bracketed(
            OneOf(
                Ref("WildcardExpressionSegment"),
                Ref("LambdaExpressionSegment"),
                Ref("BaseExpressionElementGrammar"),
            ),
        ),
    )


class LambdaExpressionSegment(BaseSegment):
    """Lambda function used in a function or columns expression.

    https://duckdb.org/docs/stable/sql/functions/lambda
    https://duckdb.org/docs/sql/expressions/star#columns-lambda-function
    """

    type = "lambda_function"
    match_grammar = OneOf(
        Sequence(
            OneOf(
                Ref("ParameterNameSegment"),
                Bracketed(Delimited(Ref("ParameterNameSegment"))),
            ),
            Ref("LambdaArrowSegment"),
            Ref("ExpressionSegment"),
        ),
        Sequence(
            "LAMBDA",
            OneOf(
                Ref("ParameterNameSegment"),
                Bracketed(Delimited(Ref("ParameterNameSegment"))),
            ),
            Ref("ColonSegment"),
            Ref("ExpressionSegment"),
        ),
    )


class ListComprehensionExpressionSegment(BaseSegment):
    """A list comprehension expression in duckdb.

    https://duckdb.org/docs/sql/functions/list#list-comprehension
    """

    type = "list_comprehension"
    match_grammar = Bracketed(
        Ref("ExpressionSegment"),
        "FOR",
        Ref("ParameterNameSegment"),
        "IN",
        Ref("ExpressionSegment"),
        Sequence("IF", Ref("ExpressionSegment"), optional=True),
        bracket_type="square",
    )


class SelectStatementSegment(ansi.SelectStatementSegment):
    """A duckdb `SELECT` statement including optional Qualify.

    https://duckdb.org/docs/sql/query_syntax/qualify
    """

    type = "select_statement"

    match_grammar = ansi.SelectStatementSegment.match_grammar.copy(
        insert=[Ref("QualifyClauseSegment", optional=True)],
        before=Ref("OrderByClauseSegment", optional=True),
        replace_terminators=True,
        terminators=[
            Ref("SetOperatorSegment"),
            Ref("WithNoSchemaBindingClauseSegment"),
            Ref("WithDataClauseSegment"),
            Sequence("ON", "CONFLICT"),
            "RETURNING",
            Ref("WithCheckOptionSegment"),
            Ref("MetaCommandQueryBufferSegment"),
        ],
    ).copy(
        # `USING SAMPLE` is query-level, so it trails the whole statement.
        insert=[Ref("UsingSampleClauseSegment", optional=True)],
    )


class UnorderedSelectStatementSegment(ansi.UnorderedSelectStatementSegment):
    """A `SELECT` statement without any ORDER clauses or later.

    This is designed for use in the context of set operations,
    for other use cases, we should use the main
    SelectStatementSegment.
    """

    type = "select_statement"

    match_grammar: Matchable = Sequence(
        OneOf(
            Sequence(
                Ref("SelectClauseSegment"),
                Ref("FromClauseSegment", optional=True),
            ),
            Sequence(
                # From-First Syntax:
                # https://duckdb.org/docs/sql/query_syntax/from
                Ref("FromClauseSegment"),
                Ref("SelectClauseSegment", optional=True),
            ),
        ),
        Ref("WhereClauseSegment", optional=True),
        Ref("GroupByClauseSegment", optional=True),
        Ref("HavingClauseSegment", optional=True),
        Ref("NamedWindowSegment", optional=True),
        Ref("QualifyClauseSegment", optional=True),
        # `USING SAMPLE` is query-level, so it also trails a `SELECT` that is a
        # branch of a set expression (e.g. one side of a `UNION`).
        Ref("UsingSampleClauseSegment", optional=True),
        terminators=[
            Ref("SetOperatorSegment"),
            Ref("OrderByClauseSegment"),
            Ref("LimitClauseSegment"),
            Sequence("ON", "CONFLICT"),
            "RETURNING",
        ],
    )


class TableReferenceSegment(ansi.TableReferenceSegment):
    """A reference to an table, CTE, subquery or alias.

    Overload for DuckDB as only tables can be single quoted identifiers
    when used by the httpfs extension.
    """

    match_grammar = ansi.TableReferenceSegment.match_grammar.copy(
        insert=[Ref("SingleQuotedIdentifierSegment")],
    )


class AliasExpressionSegment(ansi.AliasExpressionSegment):
    """A reference to an object with an `AS` clause.

    The optional AS keyword allows both implicit and explicit aliasing.
    Inherit from ANSI instead of Postgres.
    """


class OrderByClauseSegment(ansi.OrderByClauseSegment):
    """A `ORDER BY` clause like in `SELECT`."""

    match_grammar: Matchable = Sequence(
        "ORDER",
        "BY",
        Indent,
        Delimited(
            Sequence(
                OneOf(
                    "ALL",
                    Ref("ColumnReferenceSegment"),
                    Ref("NumericLiteralSegment"),
                    Ref("ExpressionSegment"),
                ),
                OneOf("ASC", "DESC", optional=True),
                Sequence("NULLS", OneOf("FIRST", "LAST"), optional=True),
            ),
            allow_trailing=True,
            terminators=[Ref("OrderByClauseTerminators")],
        ),
        Dedent,
    )


class GroupByClauseSegment(ansi.GroupByClauseSegment):
    """A `GROUP BY` clause like in `SELECT`."""

    match_grammar: Matchable = Sequence(
        "GROUP",
        "BY",
        Indent,
        Delimited(
            OneOf(
                "ALL",
                Ref("ColumnReferenceSegment"),
                Ref("NumericLiteralSegment"),
                Ref("ExpressionSegment"),
            ),
            allow_trailing=True,
            terminators=[Ref("GroupByClauseTerminatorGrammar")],
        ),
        Dedent,
    )


class QualifyClauseSegment(BaseSegment):
    """A `QUALIFY` clause like in `SELECT`.

    https://duckdb.org/docs/sql/query_syntax/qualify.html
    """

    type = "qualify_clause"
    match_grammar = Sequence(
        "QUALIFY",
        ImplicitIndent,
        OptionallyBracketed(Ref("ExpressionSegment")),
        Dedent,
    )


class ObjectLiteralElementSegment(ansi.ObjectLiteralElementSegment):
    """An object literal element segment."""

    match_grammar: Matchable = Sequence(
        OneOf(
            Ref("SingleIdentifierGrammar"),
            Ref("QuotedLiteralSegment"),
        ),
        Ref("ColonSegment"),
        Ref("BaseExpressionElementGrammar"),
    )


class StatementSegment(postgres.StatementSegment):
    """An element in the targets of a select statement."""

    match_grammar = postgres.StatementSegment.match_grammar.copy(
        insert=[
            Ref("SimplifiedPivotExpressionSegment"),
            Ref("SimplifiedUnpivotExpressionSegment"),
        ]
    )


class FromPivotExpressionSegment(BaseSegment):
    """A PIVOT expression."""

    type = "from_pivot_expression"
    match_grammar = Sequence(
        "PIVOT",
        Bracketed(
            Delimited(
                Sequence(
                    Ref("FunctionSegment"),
                    Ref("AliasExpressionSegment", optional=True),
                ),
                allow_trailing=True,
            ),
            "FOR",
            AnyNumberOf(
                Sequence(
                    Ref("SingleIdentifierGrammar"),
                    "IN",
                    Bracketed(Delimited(Ref("LiteralGrammar"), allow_trailing=True)),
                ),
            ),
            Ref("GroupByClauseSegment", optional=True),
            Ref("OrderByClauseSegment", optional=True),
            Ref("LimitClauseSegment", optional=True),
        ),
        reset_terminators=True,
    )


class SimplifiedPivotExpressionSegment(BaseSegment):
    """The DuckDB simplified PIVOT syntax.

    https://duckdb.org/docs/sql/statements/pivot#simplified-pivot-full-syntax-diagram
    """

    type = "simplified_pivot"
    match_grammar = Sequence(
        OneOf("PIVOT", "PIVOT_WIDER"),
        Ref("TableExpressionSegment"),
        Sequence(
            "ON",
            Delimited(
                OneOf(
                    Ref("ColumnReferenceSegment"),
                    Ref("ExpressionSegment"),
                ),
                Sequence(
                    "IN",
                    Bracketed(Delimited(Ref("LiteralGrammar"), allow_trailing=True)),
                    optional=True,
                ),
                allow_trailing=True,
            ),
            optional=True,
        ),
        Sequence(
            "USING",
            Delimited(
                Sequence(
                    Ref("FunctionSegment"),
                    Ref("AliasExpressionSegment", optional=True),
                ),
                allow_trailing=True,
            ),
            optional=True,
        ),
        Ref("GroupByClauseSegment", optional=True),
        Ref("OrderByClauseSegment", optional=True),
        Ref("LimitClauseSegment", optional=True),
    )


class FromUnpivotExpressionSegment(BaseSegment):
    """An UNPIVOT expression."""

    type = "from_unpivot_expression"
    match_grammar = Sequence(
        "UNPIVOT",
        Sequence("INCLUDE", "NULLS", optional=True),
        Bracketed(
            OneOf(
                Ref("SingleIdentifierGrammar"),
                Bracketed(
                    Delimited(Ref("SingleIdentifierGrammar"), allow_trailing=True)
                ),
            ),
            "FOR",
            AnyNumberOf(
                Sequence(
                    Ref("SingleIdentifierGrammar"),
                    "IN",
                    Bracketed(
                        Delimited(
                            Sequence(
                                OptionallyBracketed(
                                    Delimited(
                                        Ref("SingleIdentifierGrammar"),
                                        allow_trailing=True,
                                    )
                                ),
                                Ref("AliasExpressionSegment", optional=True),
                            ),
                            Ref("ColumnsExpressionGrammar"),
                            allow_trailing=True,
                        ),
                    ),
                ),
                min_times=1,
            ),
        ),
    )


class SimplifiedUnpivotExpressionSegment(BaseSegment):
    """The DuckDB simplified UNPIVOT syntax.

    https://duckdb.org/docs/sql/statements/unpivot#simplified-unpivot-full-syntax-diagram
    """

    type = "simplified_unpivot"
    match_grammar = Sequence(
        OneOf("UNPIVOT", "PIVOT_LONGER"),
        Ref("TableExpressionSegment"),
        "ON",
        Delimited(
            Sequence(
                OneOf(
                    Ref("ExpressionSegment"),
                    Bracketed(Delimited(Ref("ExpressionSegment"), allow_trailing=True)),
                ),
                Ref("AliasExpressionSegment", optional=True),
            ),
            allow_trailing=True,
        ),
        Sequence(
            "INTO",
            "NAME",
            Ref("SingleIdentifierGrammar"),
            "VALUE",
            Delimited(
                Ref("SingleIdentifierGrammar"),
                allow_trailing=True,
            ),
            optional=True,
        ),
        Ref("OrderByClauseSegment", optional=True),
        Ref("LimitClauseSegment", optional=True),
    )


class CreateViewStatementSegment(postgres.CreateViewStatementSegment):
    """An `Create VIEW` statement.

    https://duckdb.org/docs/sql/statements/create_view.html
    """

    type = "create_view_statement"

    match_grammar = Sequence(
        "CREATE",
        Ref("OrReplaceGrammar", optional=True),
        Ref("TemporaryGrammar", optional=True),
        "VIEW",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        Ref("BracketedColumnReferenceListGrammar", optional=True),
        "AS",
        OneOf(
            OptionallyBracketed(Ref("SelectableGrammar")),
            Ref("ValuesClauseSegment"),
        ),
    )


class CreateFunctionStatementSegment(postgres.CreateFunctionStatementSegment):
    """A `CREATE MACRO` or `CREATE FUNCTION` statement.

    https://duckdb.org/docs/sql/statements/create_macro
    """

    match_grammar = Sequence(
        "CREATE",
        Ref("OrReplaceGrammar", optional=True),
        Ref("TemporaryGrammar", optional=True),
        OneOf("MACRO", "FUNCTION"),
        Ref("FunctionNameSegment"),
        Ref("FunctionParameterListGrammar"),
        "AS",
        OneOf(
            Sequence("TABLE", Indent, Ref("SelectableGrammar"), Dedent),
            Ref("ExpressionSegment"),
        ),
    )


class DropFunctionStatementSegment(postgres.DropFunctionStatementSegment):
    """A `DROP MACRO` or `DROP FUNCTION` statement.

    https://duckdb.org/docs/sql/statements/drop.html
    """

    match_grammar = Sequence(
        "DROP",
        OneOf("MACRO", "FUNCTION"),
        Ref.keyword("TABLE", optional=True),
        Ref("IfExistsGrammar", optional=True),
        Ref("FunctionNameSegment"),
        Ref("DropBehaviorGrammar", optional=True),
    )


class CreateTypeStatementSegment(postgres.CreateTypeStatementSegment):
    """A `CREATE TYPE` statement.

    https://duckdb.org/docs/sql/statements/create_type.html
    """

    match_grammar = Sequence(
        "CREATE",
        "TYPE",
        Ref("DatatypeSegment"),
        "AS",
        OneOf(
            Ref("DatatypeSegment"),
            Sequence("ENUM", Bracketed(Delimited(Ref("QuotedLiteralSegment")))),
            Ref("StructTypeSegment"),
            Sequence("UNION", Ref("StructTypeSchemaSegment")),
        ),
    )


class CopyStatementSegment(postgres.CopyStatementSegment):
    """A `COPY` statement.

    As specified in https://duckdb.org/docs/stable/sql/statements/copy.html
    """

    type = "copy_statement"

    _target_subset = OneOf(
        Ref("QuotedLiteralSegment"), Sequence("PROGRAM", Ref("QuotedLiteralSegment"))
    )

    _table_definition = Sequence(
        Ref("TableReferenceSegment"),
        Bracketed(Delimited(Ref("ColumnReferenceSegment")), optional=True),
    )

    # DuckDB COPY TO options
    # https://duckdb.org/docs/stable/sql/statements/copy.html
    _copy_to_option = Sequence(
        Ref.keyword("WITH", optional=True),
        Bracketed(
            Delimited(
                OneOf(
                    Sequence("FORMAT", Ref("SingleIdentifierGrammar")),
                    Sequence(
                        "OVERWRITE_OR_IGNORE",
                        Ref("BooleanLiteralGrammar", optional=True),
                    ),
                    Sequence("OVERWRITE", Ref("BooleanLiteralGrammar", optional=True)),
                    Sequence("APPEND", Ref("BooleanLiteralGrammar", optional=True)),
                    Sequence(
                        "PARTITION_BY",
                        OneOf(
                            Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                            Ref("ColumnReferenceSegment"),
                        ),
                    ),
                    Sequence(
                        "WRITE_PARTITION_COLUMNS",
                        Ref("BooleanLiteralGrammar", optional=True),
                    ),
                    Sequence(
                        "COMPRESSION",
                        OneOf(
                            Ref("QuotedLiteralSegment"),
                            Ref("SingleIdentifierGrammar"),
                        ),
                    ),
                    Sequence("COMPRESSION_LEVEL", Ref("NumericLiteralSegment")),
                    Sequence("ROW_GROUP_SIZE_BYTES", Ref("NumericLiteralSegment")),
                    Sequence("ROW_GROUP_SIZE", Ref("NumericLiteralSegment")),
                    Sequence("PARQUET_VERSION", Ref("QuotedLiteralSegment")),
                )
            )
        ),
        optional=True,
    )

    # DuckDB COPY FROM options (keeping PostgreSQL compatibility for now)
    _copy_from_option = postgres.CopyStatementSegment._option

    match_grammar = Sequence(
        "COPY",
        OneOf(
            # COPY TO
            Sequence(
                OneOf(_table_definition, Bracketed(Ref("SelectableGrammar"))),
                "TO",
                Ref("QuotedLiteralSegment"),
                _copy_to_option,
            ),
            # COPY FROM
            Sequence(
                _table_definition,
                "FROM",
                Ref("QuotedLiteralSegment"),
                _copy_from_option,
            ),
        ),
    )


class SetStatementSegment(postgres.SetStatementSegment):
    """A `SET` Statement.

    DuckDB documentation: https://duckdb.org/docs/sql/statements/set
    """

    type = "set_statement"

    match_grammar = Sequence(
        "SET",
        OneOf(
            Sequence(
                "VARIABLE",
                Ref("NakedIdentifierSegment"),  # variable_name
                OneOf("TO", Ref("EqualsSegment")),
                OneOf(
                    Ref("LiteralGrammar"),
                    Ref("NakedIdentifierSegment"),
                    Ref("QuotedIdentifierSegment"),
                ),
            ),
            Sequence(
                OneOf("SESSION", "LOCAL", "GLOBAL", optional=True),
                Ref("ParameterNameSegment"),
                OneOf("TO", Ref("EqualsSegment")),
                OneOf(
                    "DEFAULT",
                    Delimited(
                        Ref("LiteralGrammar"),
                        Ref("NakedIdentifierSegment"),
                        Ref("QuotedIdentifierSegment"),
                        Ref("OnKeywordAsIdentifierSegment"),
                    ),
                ),
            ),
        ),
    )


class ArrayLiteralSegment(BaseSegment):
    """An array literal segment.

    An unqualified array literal:
    e.g. [1, 2, 3]

    DuckDB allows for trailing commas:
    e.g. [1, 2, 3,]
    """

    type = "array_literal"
    match_grammar: Matchable = Bracketed(
        Delimited(
            Ref("BaseExpressionElementGrammar"), optional=True, allow_trailing=True
        ),
        bracket_type="square",
    )


class ValuesClauseSegment(postgres.ValuesClauseSegment):
    """A `VALUES` clause within in `WITH` or `SELECT`."""

    match_grammar = Sequence(
        "VALUES",
        Delimited(
            Bracketed(
                Delimited(
                    Ref("ExpressionSegment"),
                    # DEFAULT keyword used in
                    # INSERT INTO statement.
                    "DEFAULT",
                    allow_trailing=True,
                ),
                parse_mode=ParseMode.GREEDY,
            ),
            allow_trailing=True,
        ),
        Ref("AliasExpressionSegment", optional=True),
        Ref("OrderByClauseSegment", optional=True),
        Ref("LimitClauseSegment", optional=True),
    )



"""Discovery methods for sql files.

The main public method here is `paths_from_path` which takes
potentially ambiguous paths and file input and resolves them
into specific file references. The method also processes the
`.sqlfluffignore` functionality in the process.
"""

import logging
import os
from collections.abc import Iterable, Iterator, Sequence
from pathlib import Path
from typing import Callable, Optional, Protocol

import pathspec

from sqlfluff.core.config.file import load_config_file_as_dict
from sqlfluff.core.errors import SQLFluffUserError
from sqlfluff.core.helpers.file import iter_intermediate_paths

# Instantiate the linter logger
linter_logger: logging.Logger = logging.getLogger("sqlfluff.linter")

WalkableType = Iterable[tuple[str, Optional[list[str]], list[str]]]


class IgnoreSpec(Protocol):
    """Protocol for ignore spec implementations used by this module."""

    def match_file(self, file: str) -> object:
        """Return a truthy match result if the file is ignored."""
        ...  # pragma: no cover


IgnoreSpecRecord = tuple[str, str, IgnoreSpec]
IgnoreSpecRecords = list[IgnoreSpecRecord]


def _check_ignore_specs(
    absolute_filepath: str, ignore_specs: IgnoreSpecRecords
) -> Optional[str]:
    """Check a filepath against the loaded ignore files.

    Returns:
        The path of an ignorefile if found, None otherwise.
    """
    for dirname, filename, spec in ignore_specs:
        if spec.match_file(os.path.relpath(absolute_filepath, dirname)):
            return os.path.join(dirname, filename)
    return None


def _load_specs_from_lines(lines: Iterable[str], logging_reference: str) -> IgnoreSpec:
    """Load the ignore spec from an iterable of lines.

    Raises SQLFluffUserError if unparsable for any reason.
    """
    try:
        return pathspec.PathSpec.from_lines("gitignore", lines)
    except Exception:
        _error_msg = f"Error parsing ignore patterns in {logging_reference}"
        # If the iterable is a Sequence type, then include the patterns.
        if isinstance(lines, Sequence):
            _error_msg += f": {lines}"
        raise SQLFluffUserError(_error_msg)


def _load_ignorefile(dirpath: str, filename: str) -> IgnoreSpecRecord:
    """Load a sqlfluffignore file, returning the parsed spec."""
    filepath = os.path.join(dirpath, filename)
    with open(filepath, mode="r") as f:
        spec = _load_specs_from_lines(f, filepath)
    return dirpath, filename, spec


def _load_configfile(dirpath: str, filename: str) -> Optional[IgnoreSpecRecord]:
    """Load ignore specs from a standard config file.

    This function leverages the caching used in the config module
    to ensure that anything loaded here, can be reused later. Those
    functions also handle the difference between toml and ini based
    config files.
    """
    filepath = os.path.join(dirpath, filename)
    # Use normalised path to ensure reliable caching.
    config_dict = load_config_file_as_dict(Path(filepath).resolve())
    ignore_section = config_dict.get("core", {})
    if not isinstance(ignore_section, dict):
        return None  # pragma: no cover
    patterns = ignore_section.get("ignore_paths", [])
    # If it's already a list, then we don't need to edit `patterns`,
    # but if it's not then we either split a string into a list and
    # then process it, or if there's nothing in the patterns list
    # (or the pattern input is invalid by not being something other
    # than a string or list) then we assume there's no ignore pattern
    # to process and just return None.
    if isinstance(patterns, str):
        patterns = patterns.split(",")
    elif not patterns or not isinstance(patterns, list):
        return None
    # By reaching here, we think there is a valid set of ignore patterns
    # to process.
    spec = _load_specs_from_lines(patterns, filepath)
    return dirpath, filename, spec


ignore_file_loaders: dict[str, Callable[[str, str], Optional[IgnoreSpecRecord]]] = {
    ".sqlfluffignore": _load_ignorefile,
    "pyproject.toml": _load_configfile,
    ".sqlfluff": _load_configfile,
}


def _iter_config_files(
    target_path: Path,
    working_path: Path,
) -> Iterator[tuple[str, str]]:
    """Iterate through paths looking for valid config files."""
    for search_path in iter_intermediate_paths(target_path.absolute(), working_path):
        for _filename in ignore_file_loaders:
            filepath = os.path.join(search_path, _filename)
            if os.path.isfile(filepath):
                # Yield if a config file with this name exists at this path.
                yield str(search_path), _filename


def _match_file_extension(filepath: str, valid_extensions: Sequence[str]) -> bool:
    """Match file path against extensions.

    Assumes that valid_extensions is already all lowercase.

    Returns:
        True if the file has an extension in `valid_extensions`.
    """
    filepath = filepath.lower()
    return any(filepath.endswith(ext) for ext in valid_extensions)


def _process_exact_path(
    path: str,
    working_path: str,
    lower_file_exts: tuple[str, ...],
    outer_ignore_specs: IgnoreSpecRecords,
) -> list[str]:
    """Handle exact paths being passed to paths_from_path.

    If it's got the right extension and it's not ignored, then
    we just return the normalised version of the path. If it's
    not the right extension, return nothing, and if it's ignored
    then return nothing, but include a warning for the user.
    """
    # Does it have a relevant extension? If not, just return an empty list.
    if not _match_file_extension(path, lower_file_exts):
        return []

    # It's an exact file. We only need to handle the outer ignore files.
    # There won't be any "inner" ignores because an exact file doesn't create
    # any sub paths.
    abs_fpath = os.path.abspath(path)
    ignore_file = _check_ignore_specs(abs_fpath, outer_ignore_specs)

    if not ignore_file:
        # If not ignored, just return the file.
        return [os.path.normpath(path)]

    ignore_rel_path = os.path.relpath(ignore_file, working_path)
    linter_logger.warning(
        f"Exact file path {path} was given but it was "
        f"ignored by an ignore pattern set in {ignore_rel_path}, "
        "re-run with `--disregard-sqlfluffignores` to not process "
        "ignore files."
    )
    # Return no match, because the file is ignored.
    return []


def _iter_files_in_path(
    path: str,
    ignore_files: bool,
    outer_ignore_specs: IgnoreSpecRecords,
    lower_file_exts: tuple[str, ...],
) -> Iterator[str]:
    """Handle directory paths being passed to paths_from_path.

    We're going to walk the path progressively, processing ignore
    files as we go. Those ignore files that we find (inner ignore
    files) only apply within the folder they are found, whereas the
    ignore files from outside the path (the outer ignore files) will
    always apply, so we handle them separately.
    """
    inner_ignore_specs: IgnoreSpecRecords = []
    ignore_filename_set = frozenset(ignore_file_loaders.keys())

    for dirname, subdirs, filenames in os.walk(path, topdown=True):
        # Before adding new ignore specs, remove any which are no longer relevant
        # as indicated by us no longer being in a subdirectory of them.
        # NOTE: Slice so we can modify as we go.
        for inner_dirname, inner_file, inner_spec in inner_ignore_specs[:]:
            if not (
                dirname == inner_dirname
                or dirname.startswith(os.path.abspath(inner_dirname) + os.sep)
            ):
                inner_ignore_specs.remove((inner_dirname, inner_file, inner_spec))

        # Then look for any ignore files in the path (if ignoring files), add them
        # to the inner buffer if found.
        if ignore_files:
            for ignore_file in set(filenames) & ignore_filename_set:
                ignore_spec = ignore_file_loaders[ignore_file](dirname, ignore_file)
                if ignore_spec:
                    inner_ignore_specs.append(ignore_spec)

        # Then prune any subdirectories which are ignored (by modifying `subdirs`)
        # https://docs.python.org/3/library/os.html#os.walk
        for subdir in subdirs[:]:  # slice it so that we can modify it in the process.
            # NOTE: The "*" in this next section is a bit of a hack, but pathspec
            # doesn't like matching _directories_ directly, but if we instead match
            # `directory/*` we get the same effect.
            absolute_path = os.path.abspath(os.path.join(dirname, subdir, "*"))
            if _check_ignore_specs(
                absolute_path, outer_ignore_specs
            ) or _check_ignore_specs(absolute_path, inner_ignore_specs):
                subdirs.remove(subdir)
                continue

        # Then look for any relevant sql files in the path.
        for filename in filenames:
            relative_path = os.path.join(dirname, filename)
            absolute_path = os.path.abspath(relative_path)

            # Check file extension is relevant
            if not _match_file_extension(filename, lower_file_exts):
                continue
            # Check not ignored by outer & inner ignore specs
            if _check_ignore_specs(absolute_path, outer_ignore_specs):
                continue
            if _check_ignore_specs(absolute_path, inner_ignore_specs):
                continue

            # If we get here, it's one we want. Yield it.
            yield os.path.normpath(relative_path)


def paths_from_path(
    path: str,
    ignore_non_existent_files: bool = False,
    ignore_files: bool = True,
    working_path: str = os.getcwd(),
    target_file_exts: Sequence[str] = (".sql",),
    check_non_existent_file: bool = False,
) -> list[str]:
    """Return a set of sql file paths from a potentially more ambiguous path string.

    Here we also deal with the any ignore files file if present, whether as raw
    ignore files (`.sqlfluffignore`) or embedded in more general config files like
    `.sqlfluff` or `pyproject.toml`.

    Only files within the path provided are returned, *however* the search area
    for ignore files is wider. They can both be within the provided path, and also
    between the working path and the given path.

    NOTE: In the situation that the given path is *not* a subdirectory of the
    working path, the current behaviour is to search for the *lowest common path*
    of the two. This might be counterintuitive, but supports an appropriate solution
    for the dbt templater without having to additionally pass the project root path.
    """
    if not os.path.exists(path) and not check_non_existent_file:
        if ignore_non_existent_files:
            return []
        else:
            raise SQLFluffUserError(
                f"Specified path does not exist. Check it/they exist(s): {path}."
            )

    lower_file_exts = tuple(ext.lower() for ext in target_file_exts)

    # First load any ignore files from outside the path.
    # These will be applied to every file within the path, because we know that
    # they're in a parent folder.
    outer_ignore_specs: IgnoreSpecRecords = []
    # Only load them if we're using ignore files. NOTE: That if `ignore_files`
    # is False, we keep the routines for _checking_ we just never load the
    # files in the first place.
    if ignore_files:
        for ignore_path, ignore_file in _iter_config_files(
            Path(path).absolute(),
            Path(working_path) if isinstance(working_path, str) else working_path,
        ):
            ignore_spec = ignore_file_loaders[ignore_file](ignore_path, ignore_file)
            if ignore_spec:
                outer_ignore_specs.append(ignore_spec)

    # Handle being passed an exact file first.
    if os.path.isfile(path) or check_non_existent_file:
        return _process_exact_path(
            path, working_path, lower_file_exts, outer_ignore_specs
        )

    # Otherwise, it's not an exact path and we're going to walk the path
    # progressively, processing ignore files as we go.
    return sorted(
        _iter_files_in_path(path, ignore_files, outer_ignore_specs, lower_file_exts)
    )
